import threading
from datetime import datetime

from SapphireQuant.Channel.CtpChannel import thostmduserapi as api
from SapphireQuant.Channel.CtpChannel.CtpConvert import CtpConvert
from SapphireQuant.Core.Enum import EnumQuoteEventType


class CtpQuoteChannel(api.CThostFtdcMdSpi):
    """
    CTP行情接口
    """

    def __init__(self, user_id, password, broker_id, auth_code, app_id='client_QiPlatform_1'):
        self._on_start_event = []
        self._on_end_event = []
        self._on_bar_event = []
        self._on_tick_event = []
        self._on_day_begin_event = []
        self._on_day_end_event = []
        self._user_id = user_id
        self._password = password
        self._broker_id = broker_id
        self._app_id = app_id
        self._auth_code = auth_code
        self._md_api = None
        self._trade_spi = None
        self._front_address = None
        self._flow_path = None
        self.is_connected = False
        self.trading_date = None
        self._debug_mode = False
        self._event = threading.Event()
        self._event.clear()

        self._is_front_connected = False
        self._is_login = False

        api.CThostFtdcMdSpi.__init__(self)

    def connect(self, front_address, flow_path=None):
        """
        连接前置
        :param front_address:
        :param flow_path:
        """
        print('开始连接期货行情服务器......')
        # print(threading.currentThread().ident)
        if self.is_connected:
            print('行情服务器已连接......')
            return
        self._front_address = front_address
        self._flow_path = flow_path
        if self._flow_path is not None:
            self._md_api = api.CThostFtdcMdApi_CreateFtdcMdApi(self._flow_path)
        else:
            self._md_api = api.CThostFtdcMdApi_CreateFtdcMdApi()
        self._md_api.RegisterFront(self._front_address)
        self._md_api.RegisterSpi(self)  # 注册事件类
        # 连接前置服务器
        front_flag = self._connect_front()
        if not front_flag:
            return False

        # 登录
        login_flag = self._req_login()
        if not login_flag:
            return False
        self.is_connected = True
        print('行情句柄初始化完毕......')
        return True

    def _connect_front(self):
        self._is_front_connected = False
        self._md_api.Init()
        self._event.clear()
        if self._event.wait(10):
            return self._is_front_connected

        print('前置行情服务器连接超时.....')
        return False

    def _req_login(self):
        self._is_login = False
        login_field = api.CThostFtdcReqUserLoginField()
        login_field.BrokerID = self._broker_id
        login_field.UserID = self._user_id
        login_field.Password = self._password
        login_field.UserProductInfo = "sapphire"
        print('请求登录服务器......')
        self._md_api.ReqUserLogin(login_field, 0)
        self._event.clear()
        if self._event.wait(5):
            return self._is_login

        print('登录服务器超时.....')
        return False

    def dis_connect(self):
        """
        断开连接
        """
        print('开始断开期货行情服务器......')
        logout_field = api.CThostFtdcUserLogoutField()
        logout_field.BrokerID = self._broker_id
        logout_field.UserID = self._user_id
        print('请求登出服务器......')
        rtn = self._md_api.ReqUserLogout(logout_field, 0)
        if rtn != 0:
            print("发送登出请求失败:{0}".format(rtn))

    def OnFrontConnected(self):
        """
        连接交易服务器回调函数
        """
        print("连接行情服务器成功回调......")
        self._is_front_connected = True
        self._event.set()

    def OnFrontDisconnected(self, nReason):
        """
        交易前置服务器断开连接响应
        :param nReason:
        """
        print('行情服务器断开:{0}'.format(nReason))

    def OnRspUserLogin(self, pRspUserLogin, pRspInfo, nRequestID, bIsLast):
        """
        用户登录回调函数
        :param pRspUserLogin:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        """
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('登录失败:{0}'.format(pRspInfo.ErrorMsg))
            self._is_login = False
            self._event.set()
            return
        if bIsLast:
            try:
                self.trading_date = datetime.strptime(pRspUserLogin.TradingDay, '%Y%m%d')
                print('登录行情服务器成功......')
                print('从CTP行情系统获得交易日信息【{0}】'.format(pRspUserLogin.TradingDay))
                print('从CTP行情系统获得FrontId:{0},BrokerId:{1},UserId:{2},LoginTime:{3}'.format(pRspUserLogin.FrontID, pRspUserLogin.BrokerID, pRspUserLogin.UserID, pRspUserLogin.LoginTime))
                print('中金所时间:{0},郑商所时间:{1},大商所时间:{2},上期所时间:{3}'.format(pRspUserLogin.FFEXTime, pRspUserLogin.CZCETime, pRspUserLogin.DCETime, pRspUserLogin.SHFETime))
                # OnConnectedEvent?.BeginInvoke(null, null);
            except Exception as e:
                print(str(e))
            self._is_login = True
            self._event.set()

    def OnRspUserLogout(self, pUserLogout, pRspInfo, nRequestID, bIsLast):
        """
        客户端登出响应
        :param pUserLogout:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        :return:
        """
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('登出失败:{0}'.format(pRspInfo.ErrorMsg))
            return
        if bIsLast:
            self.is_connected = False
            print('登出成功:BrokerID:{0},UserID:{1}'.format(pUserLogout.BrokerID, pUserLogout.UserID))
            # OnDisconnectedEvent?.BeginInvoke(null, null);
            print('释放行情句柄......')
            self._md_api.Release()

    def OnRspSubMarketData(self, pSpecificInstrument, pRspInfo, nRequestID, bIsLast):
        """

        :param pSpecificInstrument:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        :return:
        """
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('订阅失败:{0}'.format(pRspInfo.ErrorMsg))
            return

        if bIsLast:
            pass

        print('订阅成功:{0}'.format(pSpecificInstrument.InstrumentID))

    def OnRspUnSubMarketData(self, pSpecificInstrument, pRspInfo, nRequestID, bIsLast):
        """

        :param pSpecificInstrument:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        :return:
        """
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('退订失败:{0}'.format(pRspInfo.ErrorMsg))
            return

        if bIsLast:
            pass

        print('退订成功:{0}'.format(pSpecificInstrument.InstrumentID))

    def OnRspError(self, pRspInfo, nRequestID, bIsLast):
        """

        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        """
        print('OnRspError:{0},ID:{1}'.format(pRspInfo.ErrorMsg, pRspInfo.ErrorID))

    def OnHeartBeatWarning(self, nTimeLapse):
        """
        心跳
        :param nTimeLapse:
        """
        print('心跳:{0}'.format(str(nTimeLapse)))

    def OnRtnDepthMarketData(self, pDepthMarketData):
        """
        行情回报
        :param pDepthMarketData:
        :return:
        """
        tick = CtpConvert.convert_tick(pDepthMarketData)
        if abs(tick.open_price) < 0.0000001:
            print('非法Tick(开盘价=0):{0}'.format(tick.to_string()))
            return
        if abs(tick.volume) < 0.0000001:
            print('非法Tick(成交量=0):{0}'.format(tick.to_string()))
            return
        if abs(tick.turnover) < 0.0000001:
            print('非法Tick(成交额=0):{0}'.format(tick.to_string()))
            return
        # print(tick.to_string())
        try:
            for event in self._on_tick_event:
                event(tick)
        except Exception as e:
            print('实盘行情模块：[on_tick]事件异常:{0},Id:{1}'.format(str(e), tick.instrument_id))
            if self._debug_mode:
                raise e

    def subscribe(self, sub_id):
        """
        订阅
        :param sub_id:
        """
        str_array = ";".join(map(str, sub_id))
        print('订阅:{0}个合约:{1}'.format(len(sub_id), str_array))
        self._md_api.SubscribeMarketData([id.encode('utf-8') for id in sub_id], len(sub_id))

    def un_subscribe(self, sub_id):
        """
        退订
        :param sub_id:
        """
        str_array = ";".join(map(str, sub_id))
        print('退订:{0}个合约:{1}'.format(len(sub_id), str_array))
        self._md_api.UnSubscribeMarketData([id.encode('utf-8') for id in sub_id], len(sub_id))

    def register_event(self, event, event_type):
        """
        注册事件
        :param event_type:
        :param event:EnumQuoteEventType
        """
        if event_type == EnumQuoteEventType.on_start:
            self._on_start_event.append(event)
        elif event_type == EnumQuoteEventType.on_end:
            self._on_end_event.append(event)
        elif event_type == EnumQuoteEventType.on_day_begin:
            self._on_day_begin_event.append(event)
        elif event_type == EnumQuoteEventType.on_day_end:
            self._on_day_end_event.append(event)
        elif event_type == EnumQuoteEventType.on_bar:
            self._on_bar_event.append(event)
        elif event_type == EnumQuoteEventType.on_tick:
            self._on_tick_event.append(event)



import datetime

from SapphireQuant.Core.CSharpUtils import CSharpUtils


class OffsetCount:
    """
    偏移量和个数
    """
    def __init__(self, offset, count):
        self._offset = offset
        self._count = count

    @property
    def offset(self):
        """
        偏移量
        :return:
        """
        return self._offset

    @property
    def count(self):
        """
        个数
        :return:
        """
        return self._count

    def to_string(self):
        """

        :return:
        """
        msg = 'offset:' + str(self._offset)
        msg += 'count:' + str(self._count)
        return msg


class DateTimeOffsetMap:
    """
    时间索引
    """
    def __init__(self, file_header):
        self._bucket_size = 32
        self._mask = self._bucket_size - 1
        self._default_key = datetime.datetime(2999, 1, 1)
        self._keys = []
        self._values = []
        for i in range(32):
            self._keys.append(datetime.datetime(2999, 1, 1))
            self._values.append(OffsetCount(999, 999))
        self._file_header = file_header

    def add(self, date_time, offset):
        """
        添加
        :param date_time:
        :param offset:
        """
        if date_time == self._default_key:
            raise Exception('Invalid Date Time')

        key = datetime.datetime(date_time.year, date_time.month, date_time.day)
        key_hash = self.hash_key(key)
        index = key_hash & self._mask  # _mask二进制0b11111,这里为了保证index在1-31
        found = False
        old_index = index
        while self._keys[index] != self._default_key:
            if self._keys[index] == key:
                found = True
                break

            key_hash += 1
            index = key_hash & self._mask
            if old_index == index:
                raise Exception("Unexpected Collision Is Detected In DateTimeOffsetMap")

        if found:
            old = self._values[index]
            self._values[index] = OffsetCount(old.offset, old.count + 1)
        else:
            self._keys[index] = key
            self._values[index] = OffsetCount(offset, 1)
            self._file_header.mark_as_dirty(True)

        return not found

    def get_trading_day_offsets(self, begin_trading_day, end_trading_day):
        """

        :param begin_trading_day:
        :param end_trading_day:
        :return:
        """
        start = begin_trading_day
        end = end_trading_day
        while True:
            data = self.try_get(start)
            if data[0]:
                start_offset = data[1]
                start_count = data[2]
                break
            start = start + datetime.timedelta(days=1)
            if start > end_trading_day:
                start_offset = -1
                end_offset = -1
                start_count = 0
                end_count = 0
                return False, start_offset, end_offset, start_count, end_count

        while True:
            data = self.try_get(end)
            if data[0]:
                end_offset = data[1]
                end_count = data[2]
                break
            end = start + datetime.timedelta(days=-1)
            if end < begin_trading_day:
                start_offset = -1
                end_offset = -1
                start_count = 0
                end_count = 0
                return False, start_offset, end_offset, start_count, end_count

        return start_offset <= end_offset, start_offset, end_offset, start_count, end_count

    def get_start_offset(self, key):
        """

        :param key:
        :return:
        """
        if key == self._default_key:
            raise Exception("Invalid Date Time")

        key = datetime.datetime(key.year, key.month, key.day)
        temp_key = self._default_key
        index = -1

        key_hash = self.hash_key(key)
        idx = key_hash & self._mask
        for i in range(0, self._bucket_size):
            if self._keys[idx] != self._default_key:
                if key == self._keys[idx]:
                    index = idx
                    break

                if key < self._keys[idx]:
                    if (temp_key == self._default_key) | (temp_key > self._keys[idx]):
                        temp_key = self._keys[idx]
                        index = idx

            idx += 1
            if idx == len(self._keys):
                idx = 0

        if index == -1:
            raise Exception("Invalid Date Time")

        value = self._values[index]
        offset = value.offset
        count = value.count
        return offset, count

    def get_end_offset(self, key):
        """

        :param key:
        :return:
        """
        if key == self._default_key:
            raise Exception("Invalid Date Time")

        key = datetime.datetime(key.year, key.month, key.day)
        temp_key = self._default_key
        index = -1

        key_hash = self.hash_key(key)
        idx = key_hash & self._mask
        for i in range(0, self._bucket_size):
            if self._keys[idx] != self._default_key:
                if key == self._keys[idx]:
                    index = idx
                    break

                if key > self._keys[idx]:
                    if (temp_key == self._default_key) or (temp_key < self._keys[idx]):
                        temp_key = self._keys[idx]
                        index = idx

            idx += 1
            if idx == len(self._keys):
                idx = 0

        if index == -1:
            raise Exception("Invalid Date Time")

        value = self._values[index]
        offset = value.offset
        count = value.count
        return offset, count

    def try_get(self, key):
        """

        :param key:
        :return:
        """
        if key == self._default_key:
            raise Exception("invalid date time")

        key = datetime.datetime(key.year, key.month, key.day)
        key_hash = self.hash_key(key)
        index = key_hash & self._mask  # _mask二进制0b11111,这里为了保证index在1-31
        found = False
        old_index = index
        while self._keys[index] != self._default_key:
            if self._keys[index] == key:
                found = True
                break

            key_hash += 1
            index = key_hash & self._mask
            if old_index == index:
                raise Exception("Unexpected Collision Is Detected In DateTimeOffsetMap")

        if found:
            value = self._values[index]
            offset = value.offset
            count = value.count
        else:
            offset = -1
            count = 0

        return found, offset, count

    def write(self, writer):
        """

        :param writer:
        """
        for i in range(0, self._bucket_size):
            writer.write_int64(CSharpUtils.get_c_sharp_ticks(self._keys[i]))
        for i in range(0, self._bucket_size):
            value = self._values[i]
            writer.write_int64(value.offset)
            writer.write_int32(value.count)

    def read(self, reader):
        """

        :param reader:
        """
        self._keys = []
        for i in range(0, self._bucket_size):
            self._keys.append(CSharpUtils.convert_c_sharp_ticks_to_py_date_time(reader.read_int64()))

        self._values = []
        for i in range(0, self._bucket_size):
            self._values.append(OffsetCount(reader.read_int64(), reader.read_int32()))

    @staticmethod
    def hash_key(time):
        """

        :param time:
        :return:
        """
        return time.day



from SapphireQuant.Config.Ctp.CtpBrokerCollection import CtpBrokerCollection


class CtpBrokerManager:
    """
    CtpBroker管理器
    """
    def __init__(self):
        self.ctp_broker_collection = CtpBrokerCollection()

    def load(self, config_path):
        """
        加载
        :param config_path:
        """
        pass

    def get_servers(self, broker_id, group_name, network_type, server_type):
        """

        :param broker_id:
        :param group_name:
        :param network_type:
        :param server_type:
        :return:
        """
        return self.ctp_broker_collection.get_servers(broker_id, group_name, network_type, server_type)

    def get_servers_by_broker_id_network_server_type(self, broker_id, network_type, server_type):
        """

        :param broker_id:
        :param network_type:
        :param server_type:
        :return:
        """
        return self.ctp_broker_collection.get_servers(broker_id, network_type, server_type)

    def get_servers_by_network_server_type(self, network_type, server_type):
        """

        :param network_type:
        :param server_type:
        :return:
        """
        return self.ctp_broker_collection.get_servers(network_type, server_type)

    def get_broker(self, broker_id):
        """

        :param broker_id:
        :return:
        """
        return self.ctp_broker_collection.get_broker(broker_id)

    def get_broker_ids(self):
        """

        :return:
        """
        return self.ctp_broker_collection.get_broker_ids()



class Quote:
    """
    行情
    """
    def __init__(self):
        self._askPrice = [0.0] * 10
        self._askVolume = [0] * 10
        self._bidPrice = [0.0] * 10
        self._bidVolume = [0] * 10

        self.bid_num_order1 = 0
        self.bid_num_order2 = 0
        self.bid_num_order3 = 0
        self.bid_num_order4 = 0
        self.bid_num_order5 = 0
        self.bid_num_order6 = 0
        self.bid_num_order7 = 0
        self.bid_num_order8 = 0
        self.bid_num_order9 = 0
        self.bid_num_order10 = 0

        self.ask_num_order1 = 0
        self.ask_num_order2 = 0
        self.ask_num_order3 = 0
        self.ask_num_order4 = 0
        self.ask_num_order5 = 0
        self.ask_num_order6 = 0
        self.ask_num_order7 = 0
        self.ask_num_order8 = 0
        self.ask_num_order9 = 0
        self.ask_num_order10 = 0

    @property
    def ask_price(self):
        return self._askPrice

    @property
    def ask_volume(self):
        return self._askVolume

    @property
    def bid_price(self):
        return self._bidPrice

    @property
    def bid_volume(self):
        return self._bidVolume

    @property
    def ask_price1(self):
        return self.ask_price[0]

    @ask_price1.setter
    def ask_price1(self, value):
        self.ask_price[0] = value

    @property
    def ask_price2(self):
        return self.ask_price[1]

    @ask_price2.setter
    def ask_price2(self, value):
        self.ask_price[1] = value

    @property
    def ask_price3(self):
        return self.ask_price[2]

    @ask_price3.setter
    def ask_price3(self, value):
        self.ask_price[2] = value

    @property
    def ask_price4(self):
        return self.ask_price[3]

    @ask_price4.setter
    def ask_price4(self, value):
        self.ask_price[3] = value

    @property
    def ask_price5(self):
        return self.ask_price[4]

    @ask_price5.setter
    def ask_price5(self, value):
        self.ask_price[4] = value

    @property
    def ask_price6(self):
        return self.ask_price[5]

    @ask_price6.setter
    def ask_price6(self, value):
        self.ask_price[5] = value

    @property
    def ask_price7(self):
        return self.ask_price[6]

    @ask_price7.setter
    def ask_price7(self, value):
        self.ask_price[6] = value

    @property
    def ask_price8(self):
        return self.ask_price[7]

    @ask_price8.setter
    def ask_price8(self, value):
        self.ask_price[7] = value

    @property
    def ask_price9(self):
        return self.ask_price[8]

    @ask_price9.setter
    def ask_price9(self, value):
        self.ask_price[8] = value

    @property
    def ask_price10(self):
        return self.ask_price[9]

    @ask_price10.setter
    def ask_price10(self, value):
        self.ask_price[9] = value

    @property
    def ask_volume1(self):
        return self.ask_volume[0]

    @ask_volume1.setter
    def ask_volume1(self, value):
        self.ask_volume[0] = value

    @property
    def ask_volume2(self):
        return self.ask_volume[1]

    @ask_volume2.setter
    def ask_volume2(self, value):
        self.ask_volume[1] = value

    @property
    def ask_volume3(self):
        return self.ask_volume[2]

    @ask_volume3.setter
    def ask_volume3(self, value):
        self.ask_volume[2] = value

    @property
    def ask_volume4(self):
        return self.ask_volume[3]

    @ask_volume4.setter
    def ask_volume4(self, value):
        self.ask_volume[3] = value

    @property
    def ask_volume5(self):
        return self.ask_volume[4]

    @ask_volume5.setter
    def ask_volume5(self, value):
        self.ask_volume[4] = value

    @property
    def ask_volume6(self):
        return self.ask_volume[5]

    @ask_volume6.setter
    def ask_volume6(self, value):
        self.ask_volume[5] = value

    @property
    def ask_volume7(self):
        return self.ask_volume[6]

    @ask_volume7.setter
    def ask_volume7(self, value):
        self.ask_volume[6] = value

    @property
    def ask_volume8(self):
        return self.ask_volume[7]

    @ask_volume8.setter
    def ask_volume8(self, value):
        self.ask_volume[7] = value

    @property
    def ask_volume9(self):
        return self.ask_volume[8]

    @ask_volume9.setter
    def ask_volume9(self, value):
        self.ask_volume[8] = value

    @property
    def ask_volume10(self):
        return self.ask_volume[9]

    @ask_volume10.setter
    def ask_volume10(self, value):
        self.ask_volume[9] = value

    @property
    def bid_price1(self):
        return self.bid_price[0]

    @bid_price1.setter
    def bid_price1(self, value):
        self.bid_price[0] = value

    @property
    def bid_price2(self):
        return self.bid_price[1]

    @bid_price2.setter
    def bid_price2(self, value):
        self.bid_price[1] = value

    @property
    def bid_price3(self):
        return self.bid_price[2]

    @bid_price3.setter
    def bid_price3(self, value):
        self.bid_price[2] = value

    @property
    def bid_price4(self):
        return self.bid_price[3]

    @bid_price4.setter
    def bid_price4(self, value):
        self.bid_price[3] = value

    @property
    def bid_price5(self):
        return self.bid_price[4]

    @bid_price5.setter
    def bid_price5(self, value):
        self.bid_price[4] = value

    @property
    def bid_price6(self):
        return self.bid_price[5]

    @bid_price6.setter
    def bid_price6(self, value):
        self.bid_price[5] = value

    @property
    def bid_price7(self):
        return self.bid_price[6]

    @bid_price7.setter
    def bid_price7(self, value):
        self.bid_price[6] = value

    @property
    def bid_price8(self):
        return self.bid_price[7]

    @bid_price8.setter
    def bid_price8(self, value):
        self.bid_price[7] = value

    @property
    def bid_price9(self):
        return self.bid_price[8]

    @bid_price9.setter
    def bid_price9(self, value):
        self.bid_price[8] = value

    @property
    def bid_price10(self):
        return self.bid_price[9]

    @bid_price10.setter
    def bid_price10(self, value):
        self.bid_price[9] = value

    @property
    def bid_volume1(self):
        return self.bid_volume[0]

    @bid_volume1.setter
    def bid_volume1(self, value):
        self.bid_volume[0] = value

    @property
    def bid_volume2(self):
        return self.bid_volume[1]

    @bid_volume2.setter
    def bid_volume2(self, value):
        self.bid_volume[1] = value

    @property
    def bid_volume3(self):
        return self.bid_volume[2]

    @bid_volume3.setter
    def bid_volume3(self, value):
        self.bid_volume[2] = value

    @property
    def bid_volume4(self):
        return self.bid_volume[3]

    @bid_volume4.setter
    def bid_volume4(self, value):
        self.bid_volume[3] = value

    @property
    def bid_volume5(self):
        return self.bid_volume[4]

    @bid_volume5.setter
    def bid_volume5(self, value):
        self.bid_volume[4] = value

    @property
    def bid_volume6(self):
        return self.bid_volume[5]

    @bid_volume6.setter
    def bid_volume6(self, value):
        self.bid_volume[5] = value

    @property
    def bid_volume7(self):
        return self.bid_volume[6]

    @bid_volume7.setter
    def bid_volume7(self, value):
        self.bid_volume[6] = value

    @property
    def bid_volume8(self):
        return self.bid_volume[7]

    @bid_volume8.setter
    def bid_volume8(self, value):
        self.bid_volume[7] = value

    @property
    def bid_volume9(self):
        return self.bid_volume[8]

    @bid_volume9.setter
    def bid_volume9(self, value):
        self.bid_volume[8] = value

    @property
    def bid_volume10(self):
        return self.bid_volume[9]

    @bid_volume10.setter
    def bid_volume10(self, value):
        self.bid_volume[9] = value

    def clone(self):
        quote = Quote()
        for data in self._askPrice:
            quote._askPrice.append(data)
        for data in self._askVolume:
            quote._askVolume.append(data)
        for data in self._bidPrice:
            quote._bidPrice.append(data)
        for data in self._bidVolume:
            quote._bidVolume.append(data)

    def to_string(self):
        raise Exception("该功能未实现")


import os
import smtplib
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText


class EmailToolBox:
    """
    邮件工具箱
    """

    def __init__(self):
        pass

    @staticmethod
    def send_email(subject, msg, from_address, to_address_list, smtp_host, password, attach_path_list=None):
        """
        发送邮件
        :param subject: 邮件主题
        :param msg: 邮件内容
        :param from_address: 收信人的邮箱地址，例如：["382892414@qq.com", 'maxtrics@aliyun.com']
        :param to_address_list: 发信人的邮箱地址
        :param smtp_host: smtp服务地址，可以在邮箱看，比如163邮箱为smtp.163.com
        :param password: 发信人的邮箱密码
        :param attach_path_list: 附件地址
        """
        mail_msg = MIMEMultipart()
        mail_msg['Subject'] = subject
        mail_msg['From'] = from_address
        mail_msg['To'] = ','.join(to_address_list)
        mail_msg.attach(MIMEText(msg, 'html', 'utf-8'))

        if attach_path_list is not None:
            for attach_path in attach_path_list:
                app = MIMEApplication(open(attach_path, 'rb').read(), attach_path.split('.')[-1])
                app.add_header('Content-Disposition', 'attachment', filename=os.path.basename(attach_path))
                mail_msg.attach(app)
        try:
            server = smtplib.SMTP_SSL(smtp_host)
            server.connect(smtp_host, 465)  # 连接smtp服务器
            server.login(from_address, password)  # 登录邮箱
            server.sendmail(from_address, to_address_list, mail_msg.as_string())  # 发送邮件
            server.quit()
        except Exception as e:
            print(str(e))
            print("Error: unable to send email")





from .CtpBroker import CtpBroker
from .CtpBrokerCollection import CtpBrokerCollection
from .CtpBrokerManager import CtpBrokerManager
from .CtpCounter import CtpCounter
from .CtpCounterGroup import CtpCounterGroup
from .CtpServer import CtpServer
from .EnumNetworkType import EnumNetworkType
from .EnumServerType import EnumServerType

__all__ = [
    'CtpBroker',
    'CtpBrokerCollection',
    'CtpBrokerManager',
    'CtpCounter',
    'CtpCounterGroup',
    'CtpServer',
    'EnumNetworkType',
    'EnumServerType',
]



from datetime import datetime


class Holiday:
    """
    节假日信息
    """
    def __init__(self):
        self.name = ''
        self.pre_holiday_trading_date = datetime.min
        self.begin_date = datetime.min
        self.end_date = datetime.min
        self.post_holiday_trading_date = datetime.min

    def to_string(self):
        """

        :return:
        """
        return '{0}:[{1}-{2}],假期前最后一个交易日:{3}.假期后第一个交易日:{4}'.format(self.name,
                                                                   self.begin_date.strftime('%Y%m%d'), self.end_date.strftime('%Y%m%d'),
                                                                   self.pre_holiday_trading_date.strftime('%Y%m%d'), self.post_holiday_trading_date.strftime('%Y%m%d'))



from abc import ABC

from SapphireQuant.Core.Futures import FutureTrade

from SapphireQuant.Core.Enum import EnumBuySell, EnumMarket, EnumOrderStatus
from SapphireQuant.Simulate.Match.BaseMatchMachine import BaseMatchMachine


class OrderPriceMatchMachine(BaseMatchMachine, ABC):
    """
    委托价撮合机
    委托必成
    """

    def __init__(self):
        super().__init__()

    def cancel(self, order):
        """
        撤单
        :param order:
        :return:
        """
        if order.instrument_id in self.order_map.keys():
            orders = self.order_map[order.instrument_id]
            for i in range(len(orders)):
                data = orders[i]
                if data.local_id == order.loacl_id or data.order_sys_id == order.order_sys_id:
                    order = data
                    del orders[i]
                    break
            msg = '撤单成功'
            return True, order, msg

        msg = '未找到委托编号为{0}的订单'.format(order.order_sys_id)
        return False, order, msg

    def match_by_bar(self, bar):
        """
        撮合
        :param bar:
        :return:
        """
        lst_tuples = []
        if not self.check_bar(bar):
            return lst_tuples

        if len(self.order_map) <= 0:
            return lst_tuples

        if bar.instrument_id not in self.order_map.keys():
            return lst_tuples

        orders = self.order_map[bar.instrument_id]
        for i in range(len(orders)):
            order_item = orders[i]
            match_price = order_item.order_price
            match_volume = order_item.order_volume
            result = self.can_match_by_bar(bar, order_item)
            can_order = result[0]
            if can_order:
                match_item = FutureTrade()
                match_item.task_guid = order_item.task_guid
                # match_item.project_id
                match_item.product_code = order_item.product_code
                match_item.strategy_id = order_item.strategy_id
                match_item.local_id = order_item.local_id

                match_item.order_guid = order_item.guid
                match_item.order_sys_id = order_item.order_sys_id
                match_item.instrument_id = order_item.instrument_id
                match_item.market = EnumMarket.期货
                match_item.instrument_name = order_item.instrument_name
                match_item.exchange_id = order_item.exchange_id
                match_item.trade_volume = order_item.order_volume
                match_item.trade_price = match_price
                match_item.trade_time = order_item.order_time
                match_item.direction = order_item.direction
                match_item.open_close = order_item.open_close
                match_item.hedge_flag = order_item.hedge_flag
                self.match_serial += 1
                match_item.trade_sys_id = '{0}'.format(self.match_serial)
                match_item.trading_date = order_item.trading_date

                order_item.volume_traded = match_item.trade_volume
                order_item.match_amount += match_item.trade_price * match_item.trade_volume
                if order_item.volume_traded == order_item.order_volume:
                    order_item.order_status = EnumOrderStatus.全部成交
                else:
                    order_item.order_status = EnumOrderStatus.部分成交
                if order_item.order_volume == match_volume:
                    order_item.remark = '完全成交:成交量{0},剩余{1},价格={2}元@{3}'.format(match_volume, order_item.order_volume - match_volume, match_price,
                                                                               bar.end_time.strftime('%Y/%m%d %H:%M:%S.%f'))
                else:
                    order_item.remark = '部分成交:成交量{0},剩余{1},价格={2}元@{3}'.format(match_volume, order_item.order_volume - match_volume, match_price,
                                                                               bar.end_time.strftime('%Y/%m%d %H:%M:%S.%f'))

                lst_tuples.append([order_item, match_item])
                if order_item.order_volume - order_item.volume_traded <= 0:
                    del orders[i]
                    i -= 1

        return lst_tuples

    def can_match_by_bar(self, bar, order):
        """

        :param bar:
        :param order:
        :return:
        """
        return True, order.order_price, order.order_volume

    def can_match_by_tick(self, tick, order):
        """

        :param tick:
        :param order:
        :return:
        """
        return True, order.order_price, order.order_volume



import datetime


class DateTimeSlice:
    """
    日期时间间隔
    """
    def __init__(self):
        self.__begin = datetime.datetime
        self.__end = datetime.datetime

    @property
    def begin_time(self):
        return self.__begin

    @begin_time.setter
    def begin_time(self, value):
        self.__begin = value

    @property
    def end_time(self):
        return self.__end

    @end_time.setter
    def end_time(self, value):
        self.__end = value

    @property
    def duration(self):
        return self.__end - self.__begin




class FutureProduct:
    """
    期货产品
    """
    def __init__(self, product_id="", product_name="", exchange_id=""):
        self._product_id = product_id
        self._product_name = product_name
        self._exchange_id = exchange_id
        self._futures = []
        self.future_map = {}
        self._all_slice = []

    @property
    def product_id(self):
        """
        产品id
        :return:
        """
        return self._product_id

    @product_id.setter
    def product_id(self, value):
        self._product_id = value

    @property
    def product_name(self):
        """
        产品名称
        :return:
        """
        return self._product_name

    @product_name.setter
    def product_name(self, value):
        self._product_name = value

    @property
    def exchange_id(self):
        """
        交易市场id
        :return:
        """
        return self._exchange_id

    @exchange_id.setter
    def exchange_id(self, value):
        self._exchange_id = value

    @property
    def futures(self):
        """
        期货合约列表
        :return:
        """
        return self._futures

    @futures.setter
    def futures(self, value):
        self._futures = value

    @property
    def all_slice(self):
        """

        :return:
        """
        return self._all_slice

    def init(self, product):
        """
        构造函数
        :param product:
        """
        self._product_id = product.product_id
        self._product_name = product.product_name
        self._exchange_id = product.exchange_id
        self._all_slice.append(product.all_slice)

    def to_string(self):
        """
        转成string
        :return:
        """
        string_builder = self._product_id
        string_builder += ("[" + self._product_name + "]")
        string_builder += "交易所=" + self._exchange_id

        return string_builder



from .Bar import Bar
from .BarSeries import BarSeries
from .CSharpUtils import CSharpUtils
from .DateTimeSlice import DateTimeSlice
from .Entity import Entity
from .Exchange import Exchange
from .Holiday import Holiday
from .Instrument import Instrument
from .QiList import QiList
from ..Core import QiLogger
from .Quote import Quote
from .Tick import Tick
from .TickSeries import TickSeries
from .TimeSlice import TimeSlice

from .Enum import *
from .Futures import *

__all__ = ['Bar',
           'BarSeries',
           'CSharpUtils',
           'DateTimeSlice',
           'Entity',
           'Exchange',
           'Holiday',
           'Instrument',
           'QiList',
           'QiLogger',
           'Quote',
           'Tick',
           'TickSeries',
           'TimeSlice',
           ]
__all__.extend(Enum.__all__)
__all__.extend(Futures.__all__)



from SapphireQuant.Core.Enum.EnumBarType import EnumBarType


class BarProviderInfo:
    """
    K线提供器信息
    """
    def __init__(self):
        self.time_slices = []
        self.offset = 0
        self.living_time = None
        self.interval = 0
        self.bar_type = EnumBarType.minute



import datetime
import math
import os
import sys

from dateutil.relativedelta import relativedelta

from SapphireQuant.Config.Instrument.InstrumentManager import InstrumentManager
from SapphireQuant.Config.TradingDay.TradingDayHelper import TradingDayHelper
from SapphireQuant.Core.Bar import Bar
from SapphireQuant.Core.Enum.EnumBarType import EnumBarType
from SapphireQuant.Core.Enum.EnumMarket import EnumMarket
from SapphireQuant.Core.Enum.EnumRestoration import EnumRestoration
from SapphireQuant.Core.TickSeries import TickSeries
from SapphireQuant.DataProcessing.BarProvider import BarProvider
from SapphireQuant.DataProcessing.BaseBarHelper import BaseBarHelper
from SapphireQuant.DataProcessing.BinaryDayBarStream import BinaryDayBarStream
from SapphireQuant.DataProcessing.BinaryMinBarStream import BinaryMinBarStream
from SapphireQuant.DataProcessing.BinaryTickStream import BinaryTickStream
from SapphireQuant.Core.YfTimeHelper import YfTimeHelper
import pandas as pd


class QiDataController:
    """
    量化投资数据句柄
    """

    def __init__(self, qi_data_directory):
        config_dir = os.path.join(sys.modules["ROOT_DIR"], 'Profiles')
        instrument_manager = InstrumentManager()
        instrument_manager.load(config_dir, EnumMarket.期货)

        self.data_directory = qi_data_directory
        self.instrument_manager = instrument_manager
        self.trading_day = qi_data_directory.trading_day

        self.__future_tick_path = qi_data_directory.future_tick
        self.__future_tick_cache_path = qi_data_directory.future_tick_cache
        self.__future_min_path = qi_data_directory.future_min
        self.__future_day_path = qi_data_directory.future_day

        self.__bar_series_map = {}
        self.__tick_series_map = {}

    def init(self, trading_date):
        """
        重新初始化句柄交易日和缓存数据
        :param trading_date:
        """
        self.trading_day = trading_date
        self.__bar_series_map.clear()
        self.__tick_series_map.clear()

    def __load_today_tick_series(self, market, instrument_id, trading_date):
        path = self.__future_tick_cache_path
        tick_series = TickSeries()
        try:
            file_path = os.path.join(path, trading_date.strftime('%Y%m%d'))
            file_path = os.path.join(file_path, instrument_id.split('.')[0] + ".tk")
            if instrument_id.lower() != 'index':
                instrument = self.instrument_manager[instrument_id.split('.')[0]]
                tick_stream = BinaryTickStream(market, instrument_id, instrument.exchange_id, file_path)
                tick_stream.read_by_count(tick_series, 0, sys.maxsize)
        except Exception as e:
            print(str(e))

        return tick_series

    def __load_history_tick_series(self, market, instrument_id, trading_date):
        path = self.__future_tick_path
        tick_series = TickSeries()
        try:
            file_path = os.path.join(path, trading_date.strftime('%Y%m%d'))
            file_path = os.path.join(file_path, instrument_id.split('.')[0] + ".tk")
            if instrument_id.lower() != 'index':
                instrument = self.instrument_manager[instrument_id.split('.')[0]]
                tick_stream = BinaryTickStream(market, instrument_id, instrument.exchange_id, file_path)
                tick_stream.read_by_count(tick_series, 0, sys.maxsize)
        except Exception as e:
            print(str(e))

        return tick_series

    def load_min_bar_series(self, market, instrument_id, *trading_dates):
        """
        根据交易日信息加载历史分钟线数据
        :param market:
        :param instrument_id:
        :param trading_dates:
        :return:
        """
        path = self.__future_min_path
        bar_series = []
        try:
            if len(trading_dates) == 1:
                trading_date = trading_dates[0]
                trading_date = datetime.datetime(trading_date.year, trading_date.month, trading_date.day)

                file_path = os.path.join(path, trading_date.strftime('%Y%m'))
                file_path = os.path.join(file_path, instrument_id.split('.')[0] + ".min")
                if instrument_id.lower() != 'index':
                    min_bar_steam = BinaryMinBarStream(market, instrument_id, file_path)
                    min_bar_steam.read_trading_day(bar_series, trading_date)

            if len(trading_dates) == 2:
                begin_date = trading_dates[0]
                end_date = trading_dates[1]
                begin_date = datetime.datetime(begin_date.year, begin_date.month, begin_date.day)
                end_date = datetime.datetime(end_date.year, end_date.month, end_date.day)

                begin_month = datetime.datetime(begin_date.year, begin_date.month, 1)
                end_month = datetime.datetime(end_date.year, end_date.month, 1) + relativedelta(months=+1) + datetime.timedelta(days=-1)
                date = begin_month
                while date <= end_month:
                    file_path = os.path.join(path, date.strftime('%Y%m'))
                    file_path = os.path.join(file_path, instrument_id.split('.')[0] + ".min")
                    min_bar_steam = BinaryMinBarStream(market, instrument_id, file_path)
                    min_bar_steam.read_trading_days(bar_series, begin_date, end_date)
                    date = date + relativedelta(months=+1)
        except Exception as e:
            print(str(e))

        return bar_series

    def load_day_bar_series(self, market, instrument_id, *trading_dates):
        """
        根据交易日信息加载历史日线数据
        :param market:市场
        :param instrument_id:合约编号
        :param trading_dates:
        :return:K线
        """
        path = self.__future_day_path
        bar_series = []
        begin_date = trading_dates[0]
        try:
            if len(trading_dates) == 1:
                trading_date = trading_dates[0]
                trading_date = datetime.datetime(trading_date.year, trading_date.month, trading_date.day)

                # 这边做了修改，由于数据问题，不得不多加载一天的历史数据，来修复日线bar.pre_close的错误问题
                pre_begin_date = TradingDayHelper.get_pre_trading_day(trading_date)
                begin_month = datetime.datetime(pre_begin_date.year, pre_begin_date.month, 1)
                end_month = datetime.datetime(trading_date.year, trading_date.month, 1) + relativedelta(months=+1) + datetime.timedelta(days=-1)
                date = begin_month
                while date <= end_month:
                    file_path = os.path.join(path, date.strftime('%Y%m'))
                    file_path = os.path.join(file_path, instrument_id.split('.')[0] + ".day")
                    day_bar_steam = BinaryDayBarStream(market, instrument_id, file_path)
                    day_bar_steam.read(bar_series, pre_begin_date, trading_date)
                    date = date + relativedelta(months=+1)

            if len(trading_dates) == 2:
                begin_date = trading_dates[0]
                end_date = trading_dates[1]
                begin_date = datetime.datetime(begin_date.year, begin_date.month, begin_date.day)
                end_date = datetime.datetime(end_date.year, end_date.month, end_date.day)

                # 这边做了修改，由于数据问题，不得不多加载一天的历史数据，来修复日线bar.pre_close的错误问题
                pre_begin_date = TradingDayHelper.get_pre_trading_day(begin_date)
                begin_month = datetime.datetime(pre_begin_date.year, pre_begin_date.month, 1)
                end_month = datetime.datetime(end_date.year, end_date.month, 1) + relativedelta(months=+1) + datetime.timedelta(days=-1)
                date = begin_month
                while date <= end_month:
                    file_path = os.path.join(path, date.strftime('%Y%m'))
                    file_path = os.path.join(file_path, instrument_id.split('.')[0] + ".day")
                    day_bar_steam = BinaryDayBarStream(market, instrument_id, file_path)
                    day_bar_steam.read(bar_series, pre_begin_date, end_date)
                    date = date + relativedelta(months=+1)
        except Exception as e:
            print(str(e))

        # 新增的数据处理来修复日线bar.pre_close的错误问题
        bar_series_processed = []
        pre_bar = None
        for bar in bar_series:
            if pre_bar is not None:
                bar.pre_close = pre_bar.close
            pre_bar = bar
            if bar.begin_time >= begin_date:
                bar_series_processed.append(bar)
        return bar_series_processed

    def load_tick_series(self, market, instrument_id, *trading_dates):
        """
        根据交易日信息加载历史Tick数据
        :param market:市场
        :param instrument_id:合约编号
        :param trading_dates:交易日
        :return:
        """
        if len(trading_dates) == 1:
            trading_date = trading_dates[0]
            trading_date = datetime.datetime(trading_date.year, trading_date.month, trading_date.day)
            return self.__load_history_tick_series(market, instrument_id, trading_date)
        else:
            begin_date = trading_dates[0]
            end_date = trading_dates[1]
            begin_date = datetime.datetime(begin_date.year, begin_date.month, begin_date.day)
            end_date = datetime.datetime(end_date.year, end_date.month, end_date.day)

            all_ticks = []
            try:
                begin_trading_date = TradingDayHelper.get_first_trading_day(begin_date)
                end_trading_date = TradingDayHelper.get_last_trading_day(end_date)
                date = begin_trading_date
                while date <= end_trading_date:
                    temp_ticks = self.__load_history_tick_series(market, instrument_id, date)
                    all_ticks.extend(temp_ticks)
                    date = TradingDayHelper.get_next_trading_day(date)
            except Exception as e:
                print(str(e))

            return all_ticks

    def load_tick_series_by_date_time(self, market, instrument_id, begin_time, end_time):
        """
        根据时间区间加载历史Tick数据
        :param market:市场
        :param instrument_id:合约编号
        :param begin_time:开始时间
        :param end_time:截止时间
        :return:
        """
        path = self.__future_tick_path
        all_ticks = []
        try:
            begin_trading_date = YfTimeHelper.get_trading_day(begin_time)
            end_trading_date = YfTimeHelper.get_trading_day(end_time)
            date = begin_trading_date
            while date <= end_trading_date:
                file_path = os.path.join(path, date.strftime("%Y%m%d"))
                file_path = os.path.join(file_path, instrument_id.split('.')[0] + ".tk")
                if instrument_id.lower() != "index":
                    instrument = self.instrument_manager[instrument_id.split('.')[0]]
                    tick_stream = BinaryTickStream(market, instrument_id, instrument.exchange_id, file_path)
                    tick_stream.read_by_time(all_ticks, begin_time, end_time)

                date = date + datetime.timedelta(days=1)
        except Exception as e:
            print(str(e))

        return all_ticks

    def load_tick_series_real(self, market, instrument_id, begin_time, end_time):
        """
        实盘根据时间区间获取Tick数据
        会根据DataHandler设置的交易日判断，将begin_time-end_time的区间切分成历史区间和当日区间
        历史区间：begin_time-昨日收盘
        当日区间:当日开盘-end_time
        若需要加载历史交易日Tick，则从历史目录加载数据
        若需要加载当日交易日Tick，则从当日目录加载数据
        :param market:市场
        :param instrument_id:合约编号
        :param begin_time:开始时间
        :param end_time:截止时间
        :return:
        """
        path = self.__future_tick_path
        all_ticks = []
        try:
            begin_trading_date = YfTimeHelper.get_trading_day(begin_time)
            last_trading_day = YfTimeHelper.get_trading_day(end_time)
            if last_trading_day == self.trading_day:
                end_trading_date = TradingDayHelper.get_pre_trading_day(last_trading_day)
            else:
                end_trading_date = last_trading_day
            date = begin_trading_date
            while date <= end_trading_date:
                file_path = os.path.join(path, date.strftime("%Y%m%d"))
                file_path = os.path.join(file_path, instrument_id.split('.')[0] + ".tk")
                if instrument_id.lower() != "index":
                    instrument = self.instrument_manager[instrument_id.split('.')[0]]
                    tick_stream = BinaryTickStream(market, instrument_id, instrument.exchange_id, file_path)
                    tick_stream.read_by_time(all_ticks, begin_time, end_time)

                date = date + datetime.timedelta(days=1)

            if end_trading_date != last_trading_day:
                tick_series = self.__load_today_tick_series(market, instrument_id, self.trading_day)
                all_ticks.extend(tick_series)
        except Exception as e:
            print(str(e))

        return all_ticks

    def load_bar_series_by_date(self, market, instrument_id, interval, bar_type, begin_date, end_date, *instrument_ids):
        """
        根据日期区间加载历史K线
        :param market:
        :param instrument_id:
        :param interval:
        :param bar_type:
        :param begin_date:
        :param end_date:
        :param instrument_ids:
        :return:
        """
        begin_date = datetime.datetime(begin_date.year, begin_date.month, begin_date.day)
        end_date = datetime.datetime(end_date.year, end_date.month, end_date.day)
        begin_trading_date = TradingDayHelper.get_first_trading_day(begin_date)
        end_trading_date = TradingDayHelper.get_last_trading_day(end_date)
        lst_trading_days = TradingDayHelper.get_trading_days(begin_trading_date, end_trading_date)
        bar_provider = BarProvider()
        bar_provider.create_bar_provider_by_date(self.instrument_manager, instrument_id, begin_trading_date, end_trading_date, interval,
                                                 bar_type, *instrument_ids)
        if bar_type == EnumBarType.second:
            for trading_day in lst_trading_days:
                tick_series = self.__load_history_tick_series(market, instrument_id, trading_day)
                for tick in tick_series:
                    bar_provider.add_tick(tick)
        elif (bar_type == EnumBarType.minute) | (bar_type == EnumBarType.hour):
            bar_series_min = self.load_min_bar_series(market, instrument_id, begin_trading_date, end_trading_date)
            for bar in bar_series_min:
                bar_provider.add_bar(bar)
        elif bar_type == EnumBarType.day:
            bar_series_day = self.load_day_bar_series(market, instrument_id, begin_trading_date, end_trading_date)
            for bar in bar_series_day:
                bar_provider.add_bar(bar)
        else:
            raise Exception("不支持的k线:" + str(bar_type))

        return bar_provider.bar_series

    def load_bar_series_by_date_time(self, market, instrument_id, interval, bar_type, begin_time, end_time, *instrument_ids):
        """
        根据时间区间加载历史K线
        :param market:
        :param instrument_id:
        :param interval:
        :param bar_type:
        :param begin_time:
        :param end_time:
        :param instrument_ids:
        :return:
        """
        begin_trading_date = YfTimeHelper.get_trading_day(begin_time)
        end_trading_date = YfTimeHelper.get_trading_day(end_time)
        lst_trading_days = TradingDayHelper.get_trading_days(begin_trading_date, end_trading_date)
        bar_provider = BarProvider()
        bar_provider.create_bar_provider_by_date_time(self.instrument_manager, instrument_id, begin_time, end_time, interval, bar_type,
                                                      *instrument_ids)
        if bar_type == EnumBarType.second:
            for trading_day in lst_trading_days:
                tick_series = self.__load_history_tick_series(market, instrument_id, trading_day)
                for tick in tick_series:
                    bar_provider.add_tick(tick)
        elif (bar_type == EnumBarType.minute) | (bar_type == EnumBarType.hour):
            bar_series_min = self.load_min_bar_series(market, instrument_id, begin_trading_date, end_trading_date)
            for bar in bar_series_min:
                bar_provider.add_bar(bar)
        elif bar_type == EnumBarType.day:
            bar_series_day = self.load_day_bar_series(market, instrument_id, begin_trading_date, end_trading_date)
            for bar in bar_series_day:
                bar_provider.add_bar(bar)
        else:
            raise Exception("不支持的k线:" + str(bar_type))

        return bar_provider.bar_series

    def load_bar_series_by_length(self, market, instrument_id, interval, bar_type, max_length, end_date, *instrument_ids):
        """
        根据结束日期加载指定回溯根数的历史K线（K线数据包含结束日期）
        :param market:
        :param instrument_id:
        :param interval:
        :param bar_type:
        :param max_length:
        :param end_date:
        :param instrument_ids:
        :return:
        """
        # 这里为了提升取数据的速度,根据时间预计算根数,不足再补
        end_trading_date = TradingDayHelper.get_last_trading_day(end_date)
        lst_date_time_slices_one_day = BaseBarHelper.create_one_day_date_time_slice(
            self.instrument_manager, instrument_id, end_trading_date, interval, bar_type, *instrument_ids)
        if len(lst_date_time_slices_one_day) > 0:
            n_one_day_count = len(lst_date_time_slices_one_day)
        else:
            n_one_day_count = 1

        n_trading_days = int(math.ceil(max_length * 1.0 / n_one_day_count))
        if bar_type == EnumBarType.day:
            n_trading_days = n_trading_days * interval
        else:
            n_trading_days = n_trading_days

        begin_trading_date = TradingDayHelper.get_pre_trading_day(end_trading_date, n_trading_days - 1)
        lst_trading_days = TradingDayHelper.get_trading_days(begin_trading_date, end_trading_date)
        bar_provider = BarProvider()
        bar_provider.create_bar_provider_by_date(
            self.instrument_manager, instrument_id, begin_trading_date, end_trading_date, interval, bar_type, *instrument_ids)
        if bar_provider.bar_type == EnumBarType.second:
            for trading_day in lst_trading_days:
                tick_series = self.__load_history_tick_series(market, instrument_id, trading_day)
                for tick in tick_series:
                    bar_provider.add_tick(tick)
        elif (bar_provider.bar_type == EnumBarType.minute) | (bar_provider.bar_type == EnumBarType.hour):
            bar_series_min = self.load_min_bar_series(market, instrument_id, begin_trading_date, end_trading_date)
            for bar in bar_series_min:
                bar_provider.add_bar(bar)
        elif bar_provider.bar_type == EnumBarType.day:
            bar_series_day = self.load_day_bar_series(market, instrument_id, begin_trading_date, end_trading_date)
            for bar in bar_series_day:
                bar_provider.add_bar(bar)
        else:
            raise Exception("不支持的k线:" + str(bar_type))

        bar_series = bar_provider.bar_series
        if len(bar_provider.bar_series) < max_length:
            trading_date = begin_trading_date
            count = 0
            while count < 5:
                count += 1
                trading_date = TradingDayHelper.get_pre_trading_day(trading_date)
                if bar_provider.bar_type == EnumBarType.second:
                    bar_provider = BarProvider()
                    bar_provider.create_bar_provider_by_trading_date(self.instrument_manager, instrument_id, trading_date, interval, bar_type, *instrument_ids)
                    tick_series = self.__load_history_tick_series(market, instrument_id, trading_date)
                    for tick in tick_series:
                        bar_provider.add_tick(tick)
                elif (bar_provider.bar_type == EnumBarType.minute) | (bar_provider.bar_type == EnumBarType.hour):
                    bar_provider = BarProvider()
                    bar_provider.create_bar_provider_by_trading_date(self.instrument_manager, instrument_id, trading_date, interval, bar_type, *instrument_ids)
                    bar_series_min = self.load_min_bar_series(market, instrument_id, trading_date)
                    for bar in bar_series_min:
                        bar_provider.add_bar(bar)
                elif bar_provider.bar_type == EnumBarType.day:
                    begin_date = TradingDayHelper.get_pre_trading_day(trading_date, interval - 1)
                    bar_provider = BarProvider()
                    bar_provider.create_bar_provider_by_trading_date(self.instrument_manager, instrument_id, trading_date, interval, bar_type, *instrument_ids)
                    bar_series_day = self.load_day_bar_series(market, instrument_id, begin_date, end_trading_date)
                    trading_date = begin_date
                    for bar in bar_series_day:
                        bar_provider.add_bar(bar)
                else:
                    raise Exception("不支持的k线:" + str(bar_type))

                bar_provider.bar_series.extend(bar_series)
                bar_series = bar_provider.bar_series

                if len(bar_series) >= max_length:
                    break

        bar_provider = BarProvider()
        next_trading_date = TradingDayHelper.get_next_trading_day(end_trading_date)
        bar_provider.create_bar_provider_by_trading_date(self.instrument_manager, instrument_id, next_trading_date, interval, bar_type, *instrument_ids)
        for bar in bar_series:
            bar_provider.bar_series.append(bar)

        remove_length = len(bar_provider.bar_series) - max_length
        if remove_length > 0:
            del bar_provider.bar_series[0:remove_length]

        return bar_provider.bar_series

    def load_night_am_pm_bar_series_by_date(self, market, instrument_id, begin_date, end_date, *instrument_ids):
        """
        根据日期区间加载历史K线
        :param market:
        :param instrument_id:
        :param begin_date:
        :param end_date:
        :param instrument_ids:
        :return:
        """
        begin_date = datetime.datetime(begin_date.year, begin_date.month, begin_date.day)
        end_date = datetime.datetime(end_date.year, end_date.month, end_date.day)
        begin_trading_date = TradingDayHelper.get_first_trading_day(begin_date)
        end_trading_date = TradingDayHelper.get_last_trading_day(end_date)
        bar_provider = BarProvider()
        bar_provider.create_night_am_pm_bar_provider_by_date(self.instrument_manager, instrument_id, begin_trading_date, end_trading_date, *instrument_ids)
        bar_series_min = self.load_min_bar_series(market, instrument_id, begin_trading_date, end_trading_date)
        for bar in bar_series_min:
            bar_provider.add_bar(bar)

        return bar_provider.bar_series

    def load_night_am_pm_bar_series_by_length(self, market, instrument_id, max_length, end_date, *instrument_ids):
        """
        根据结束日期加载指定回溯根数的历史K线（K线数据包含结束日期）
        :param market:
        :param instrument_id:
        :param max_length:
        :param end_date:
        :param instrument_ids:
        :return:
        """
        # 这里为了提升取数据的速度,根据时间预计算根数,不足再补
        end_trading_date = TradingDayHelper.get_last_trading_day(end_date)
        n_one_day_count = 2  # 默认等2 这样可以预载多一点的K线，再做切除

        n_trading_days = int(math.ceil(max_length * 1.0 / n_one_day_count))

        begin_trading_date = TradingDayHelper.get_pre_trading_day(end_trading_date, n_trading_days - 1)
        bar_provider = BarProvider()
        bar_provider.create_night_am_pm_bar_provider_by_date(
            self.instrument_manager, instrument_id, begin_trading_date, end_trading_date, *instrument_ids)
        bar_series_min = self.load_min_bar_series(market, instrument_id, begin_trading_date, end_trading_date)
        for bar in bar_series_min:
            bar_provider.add_bar(bar)

        bar_series = bar_provider.bar_series
        if len(bar_provider.bar_series) < max_length:
            trading_date = begin_trading_date
            count = 0
            while count < 5:
                count += 1
                trading_date = TradingDayHelper.get_pre_trading_day(trading_date)
                bar_provider = BarProvider()
                bar_provider.create_night_am_pm_bar_provider_by_date(self.instrument_manager, instrument_id, trading_date, trading_date, *instrument_ids)
                bar_series_min = self.load_min_bar_series(market, instrument_id, trading_date)
                for bar in bar_series_min:
                    bar_provider.add_bar(bar)

                bar_provider.bar_series.extend(bar_series)
                bar_series = bar_provider.bar_series

                if len(bar_series) >= max_length:
                    break

        bar_provider = BarProvider()
        next_trading_date = TradingDayHelper.get_next_trading_day(end_trading_date)
        bar_provider.create_night_am_pm_bar_provider_by_date(self.instrument_manager, instrument_id, next_trading_date, next_trading_date, *instrument_ids)
        for bar in bar_series:
            bar_provider.bar_series.append(bar)

        remove_length = len(bar_provider.bar_series) - max_length
        if remove_length > 0:
            del bar_provider.bar_series[0:remove_length]

        return bar_provider.bar_series

    def load_bar_series_by_length_limit_end_time(self, market, instrument_id, interval, bar_type, max_length, end_time, *instrument_ids):
        """
        根据结束时间加载指定回溯根数的历史K线（K线数据都小于等于结束时间）
        :param market:
        :param instrument_id:
        :param interval:
        :param bar_type:
        :param max_length:
        :param end_time:
        :param instrument_ids:
        :return:
        """
        # 这里为了提升取数据的速度,根据时间预计算根数,不足再补
        end_trading_date = YfTimeHelper.get_trading_day(end_time)
        end_trading_date = TradingDayHelper.get_last_trading_day(end_trading_date)
        lst_date_time_slices_one_day = BaseBarHelper.create_one_day_date_time_slice(
            self.instrument_manager, instrument_id, end_trading_date, interval, bar_type, *instrument_ids)
        if len(lst_date_time_slices_one_day) > 0:
            n_one_day_count = len(lst_date_time_slices_one_day)
        else:
            n_one_day_count = 1

        n_trading_days = int(math.ceil(max_length * 1.0 / n_one_day_count))
        if bar_type == EnumBarType.day:
            n_trading_days = n_trading_days * interval
        else:
            n_trading_days = n_trading_days

        begin_trading_date = TradingDayHelper.get_pre_trading_day(end_trading_date, n_trading_days - 1)
        lst_trading_days = TradingDayHelper.get_trading_days(begin_trading_date, end_trading_date)
        bar_provider = BarProvider()
        trading_time = self.instrument_manager.get_trading_time(begin_trading_date, instrument_id)
        begin_time = TradingDayHelper.get_natural_date_time(begin_trading_date, trading_time[0].begin_time)
        bar_provider.create_bar_provider_by_date_time(self.instrument_manager, instrument_id, begin_time, end_time, interval, bar_type,
                                                      *instrument_ids)
        if bar_provider.bar_type == EnumBarType.second:
            for trading_day in lst_trading_days:
                tick_series = self.__load_history_tick_series(market, instrument_id, trading_day)
                for tick in tick_series:
                    bar_provider.add_tick(tick)
        elif (bar_provider.bar_type == EnumBarType.minute) | (bar_provider.bar_type == EnumBarType.hour):
            bar_series_min = self.load_min_bar_series(market, instrument_id, begin_trading_date, end_trading_date)
            for bar in bar_series_min:
                bar_provider.add_bar(bar)
        elif bar_provider.bar_type == EnumBarType.day:
            bar_series_day = self.load_day_bar_series(market, instrument_id, begin_trading_date, end_trading_date)
            for bar in bar_series_day:
                bar_provider.add_bar(bar)
        else:
            raise Exception("不支持的k线:" + str(bar_type))

        bar_series = bar_provider.bar_series
        if len(bar_provider.bar_series) < max_length:
            trading_date = begin_trading_date
            count = 0
            while count < 5:
                count += 1
                trading_date = TradingDayHelper.get_pre_trading_day(trading_date)
                if bar_provider.bar_type == EnumBarType.second:
                    bar_provider = BarProvider()
                    bar_provider.create_bar_provider_by_trading_date(self.instrument_manager, instrument_id, trading_date, interval, bar_type, *instrument_ids)
                    tick_series = self.__load_history_tick_series(market, instrument_id, trading_date)
                    for tick in tick_series:
                        bar_provider.add_tick(tick)
                elif (bar_provider.bar_type == EnumBarType.minute) | (bar_provider.bar_type == EnumBarType.hour):
                    bar_provider = BarProvider()
                    bar_provider.create_bar_provider_by_trading_date(self.instrument_manager, instrument_id, trading_date, interval, bar_type, *instrument_ids)
                    bar_series_min = self.load_min_bar_series(market, instrument_id, trading_date)
                    for bar in bar_series_min:
                        bar_provider.add_bar(bar)
                elif bar_provider.bar_type == EnumBarType.day:
                    begin_date = TradingDayHelper.get_pre_trading_day(trading_date, interval - 1)
                    bar_provider = BarProvider()
                    bar_provider.create_bar_provider_by_trading_date(self.instrument_manager, instrument_id, trading_date, interval, bar_type, *instrument_ids)
                    bar_series_day = self.load_day_bar_series(market, instrument_id, begin_date, end_trading_date)
                    trading_date = begin_date
                    for bar in bar_series_day:
                        bar_provider.add_bar(bar)
                else:
                    raise Exception("不支持的k线:" + str(bar_type))

                bar_provider.bar_series.extend(bar_series)
                bar_series = bar_provider.bar_series

                if len(bar_series) >= max_length:
                    break

        bar_provider = BarProvider()
        bar_provider.create_bar_provider_by_trading_date(self.instrument_manager, instrument_id, self.trading_day, interval, bar_type, *instrument_ids)
        for bar in bar_series:
            bar_provider.bar_series.append(bar)

        remove_length = len(bar_provider.bar_series) - max_length
        if remove_length > 0:
            del bar_provider.bar_series[0:remove_length]

        return bar_provider.bar_series

    def get_tick_series_by_time(self, market, instrument_id, begin_time, end_time):
        """
        实盘根据时间区间获取Tick数据
        注意：若之前已经获取过品种的Tick数据，则直接从缓存返回Tick数据，不会根据变化的begin_time和end_time从新加载Tick数据
        :param market:市场
        :param instrument_id:合约编号
        :param begin_time:开始时间
        :param end_time:截止时间
        :return:
        """
        if instrument_id not in self.__tick_series_map.keys():
            all_ticks = self.load_tick_series_real(market, instrument_id, begin_time, end_time)
            self.__tick_series_map[instrument_id] = all_ticks
        else:
            all_ticks = self.__tick_series_map[instrument_id]
        return all_ticks

    def get_tick_series_by_length(self, market, instrument_id, end_time, length):
        """
        实盘根据时间区间获取Tick数据
        注意：
        若之前已经获取过品种的Tick数据，则直接从缓存返回Tick数据，
        若之前没有获取过品种的Tick数据，则加载Tick数据到缓存，并返回
        :param market:市场
        :param instrument_id:合约编号
        :param end_time:截止时间
        :param length:长度
        :return:
        """
        tick_series = []
        if instrument_id not in self.__tick_series_map.keys():
            lst_ticks = []
            end_trading_date = YfTimeHelper.get_trading_day(end_time)
            trading_date = TradingDayHelper.get_last_trading_day(end_trading_date)
            while len(lst_ticks) < length:
                temp_lst_ticks = lst_ticks.copy()
                temp_tick_series = self.load_tick_series(market, instrument_id, trading_date)
                lst_ticks.clear()
                for tick in temp_tick_series:
                    if tick.end_time <= end_time:
                        lst_ticks.append(tick)
                lst_ticks.extend(temp_lst_ticks)
                trading_date = TradingDayHelper.get_pre_trading_day(trading_date)
            tick_series = lst_ticks[-length:]
            self.__tick_series_map[instrument_id] = tick_series
        else:
            tick_series = self.__tick_series_map[instrument_id]
        return tick_series

    def get_bar_series_by_time(self, market, instrument_id, interval, bar_type, begin_time, end_time, restore=EnumRestoration.不复权, *instrument_ids):
        """
        实盘根据时间区间获取K线
        历史K线从历史数据直接加载组装
        当日K线根据Tick数据组装
        最后合并历史和当日K线，并返回
        注意：
        若之前已经获取过品种的K线数据，则直接从缓存返回K线，(判断条件：interval，bar_type的唯一性)
        若之前没有获取过品种的K线数据，则加载K线数据到缓存，并返回
        :param market:市场
        :param instrument_id:合约编号
        :param interval:间隔
        :param bar_type:K线类型
        :param begin_time:开始时间
        :param end_time:截止时间
        :param restore:复权方式
        :param instrument_ids:参与求交易时间交集的品种
        :return:
        """
        bar_providers = []
        if instrument_id in self.__bar_series_map.keys():
            bar_providers = self.__bar_series_map[instrument_id]
        else:
            self.__bar_series_map[instrument_id] = bar_providers

        # 已经获取过直接返回
        if len(bar_providers) > 0:
            bar_provider = next((data for data in bar_providers if (data.interval == interval) & (data.bar_type == bar_type)), None)
            if bar_provider is not None:
                return bar_provider.bar_series

        # 加载历史数据
        begin_trading_date = YfTimeHelper.get_trading_day(begin_time)
        last_trading_day = YfTimeHelper.get_trading_day(end_time)
        if last_trading_day == self.trading_day:
            end_trading_date = TradingDayHelper.get_pre_trading_day(last_trading_day)
        else:
            end_trading_date = last_trading_day
        lst_trading_days = TradingDayHelper.get_trading_days(begin_trading_date, end_trading_date)
        end_time = end_trading_date + datetime.timedelta(hours=16)  # 目前国内都是15点收盘
        if bar_type == EnumBarType.second:
            bar_provider = BarProvider()
            bar_provider.create_bar_provider_by_date_time(self.instrument_manager, instrument_id, begin_time, end_time, interval, bar_type,
                                                          *instrument_ids)
            for trading_day in lst_trading_days:
                tick_series = self.__load_history_tick_series(market, instrument_id, trading_day)
                for tick in tick_series:
                    bar_provider.add_tick(tick)
        elif (bar_type == EnumBarType.minute) | (bar_type == EnumBarType.hour):
            bar_provider = BarProvider()
            bar_provider.create_bar_provider_by_date_time(self.instrument_manager, instrument_id, begin_time, end_time, interval, bar_type,
                                                          *instrument_ids)
            bar_series_min = self.load_min_bar_series(market, instrument_id, begin_trading_date, end_trading_date)
            for bar in bar_series_min:
                bar_provider.add_bar(bar)
        elif bar_type == EnumBarType.day:
            bar_provider = BarProvider()
            bar_provider.create_bar_provider_by_date_time(self.instrument_manager, instrument_id, begin_time, end_time, interval, bar_type,
                                                          *instrument_ids)
            bar_series_day = self.load_day_bar_series(market, instrument_id, begin_trading_date, end_trading_date)
            for bar in bar_series_day:
                bar_provider.add_bar(bar)
        else:
            raise Exception("不支持的k线:" + str(bar_type))

        bar_series = bar_provider.bar_series
        bar_provider = BarProvider()
        bar_provider.create_bar_provider_by_trading_date(self.instrument_manager, instrument_id, self.trading_day, interval, bar_type, *instrument_ids)
        for bar in bar_series:
            bar_provider.bar_series.append(bar)

        # 组装当日数据
        self.__combine_today_tick(bar_provider, market, end_time)

        bar_providers.append(bar_provider)
        if (bar_provider.bar_type == EnumBarType.second) | (bar_provider.bar_type == EnumBarType.minute) | (bar_provider.bar_type == EnumBarType.hour):
            lst_remove = []
            for bar in bar_provider.bar_series:
                if bar.end_time < begin_time:
                    lst_remove.append(bar)
                else:
                    break
            for bar in lst_remove:
                bar_provider.bar_series.remove(bar)
        elif bar_provider.bar_type == EnumBarType.day:
            lst_remove = []
            for bar in bar_provider.bar_series:
                if bar.end_time < begin_trading_date:
                    lst_remove.append(bar)
                else:
                    break
            for bar in lst_remove:
                bar_provider.bar_series.remove(bar)
        else:
            raise Exception("不支持的k线:" + str(bar_type))

        return bar_provider.bar_series

    def get_bar_series_by_length(self, market, instrument_id, interval, bar_type, max_length, end_time, restore=EnumRestoration.不复权, *instrument_ids):
        """
        实盘根据回溯根数获取K线
        历史K线从历史数据直接加载组装
        当日K线根据Tick数据组装
        最后合并历史和当日K线，并返回
        注意：
        若之前已经获取过品种的K线数据，则直接从缓存返回K线，(判断条件：interval，bar_type的唯一性)
        若之前没有获取过品种的K线数据，则加载K线数据到缓存，并返回
        :param market:市场
        :param instrument_id:合约编号
        :param interval:间隔
        :param bar_type:K线类型
        :param max_length:回溯根数
        :param end_time:截止时间
        :param restore:复权方式
        :param instrument_ids:参与求交易时间交集的品种
        :return:
        """
        bar_providers = []
        if instrument_id in self.__bar_series_map.keys():
            bar_providers = self.__bar_series_map[instrument_id]
        else:
            self.__bar_series_map[instrument_id] = bar_providers

        # 已经获取过直接返回
        if len(bar_providers) > 0:
            bar_provider = next((data for data in bar_providers if (data.interval == interval) & (data.bar_type == bar_type)), None)
            if bar_provider is not None:
                return bar_provider.bar_series

        # 这里为了提升取数据的速度,根据时间预计算根数,不足再补
        last_trading_day = YfTimeHelper.get_trading_day(end_time)
        if last_trading_day == self.trading_day:
            end_trading_date = TradingDayHelper.get_pre_trading_day(last_trading_day)
        else:
            end_trading_date = last_trading_day
        lst_date_time_slices_one_day = BaseBarHelper.create_one_day_date_time_slice(self.instrument_manager, instrument_id, end_trading_date, interval,
                                                                                    bar_type,
                                                                                    *instrument_ids)
        if len(lst_date_time_slices_one_day) > 0:
            n_one_day_count = len(lst_date_time_slices_one_day)
        else:
            n_one_day_count = 1

        n_trading_days = int(math.ceil(max_length * 1.0 / n_one_day_count))
        if bar_type == EnumBarType.day:
            n_trading_days = n_trading_days * interval
        else:
            n_trading_days = n_trading_days

        begin_trading_date = TradingDayHelper.get_pre_trading_day(end_trading_date, n_trading_days - 1)
        lst_trading_days = TradingDayHelper.get_trading_days(begin_trading_date, end_trading_date)
        bar_provider = BarProvider()
        bar_provider.create_bar_provider_by_date_time(self.instrument_manager, instrument_id, begin_trading_date, end_trading_date, interval, bar_type,
                                                      *instrument_ids)
        if bar_provider.bar_type == EnumBarType.second:
            for trading_day in lst_trading_days:
                tick_series = self.__load_history_tick_series(market, instrument_id, trading_day)
                for tick in tick_series:
                    bar_provider.add_tick(tick)
        elif (bar_provider.bar_type == EnumBarType.minute) | (bar_provider.bar_type == EnumBarType.hour):
            bar_series_min = self.load_min_bar_series(market, instrument_id, begin_trading_date, end_trading_date)
            for bar in bar_series_min:
                bar_provider.add_bar(bar)
        elif bar_provider.bar_type == EnumBarType.day:
            bar_series_day = self.load_day_bar_series(market, instrument_id, begin_trading_date, end_trading_date)
            for bar in bar_series_day:
                bar_provider.add_bar(bar)
        else:
            raise Exception("不支持的k线:" + str(bar_type))

        bar_series = bar_provider.bar_series
        if len(bar_provider.bar_series) < max_length:
            trading_date = begin_trading_date
            count = 0
            while count < 5:
                count += 1
                trading_date = TradingDayHelper.get_pre_trading_day(trading_date)
                if bar_provider.bar_type == EnumBarType.second:
                    bar_provider = BarProvider()
                    bar_provider.create_bar_provider_by_trading_date(self.instrument_manager, instrument_id, trading_date, interval, bar_type, *instrument_ids)
                    tick_series = self.__load_history_tick_series(market, instrument_id, trading_date)
                    for tick in tick_series:
                        bar_provider.add_tick(tick)
                elif (bar_provider.bar_type == EnumBarType.minute) | (bar_provider.bar_type == EnumBarType.hour):
                    bar_provider = BarProvider()
                    bar_provider.create_bar_provider_by_trading_date(self.instrument_manager, instrument_id, trading_date, interval, bar_type, *instrument_ids)
                    bar_series_min = self.load_min_bar_series(market, instrument_id, trading_date)
                    for bar in bar_series_min:
                        bar_provider.add_bar(bar)
                elif bar_provider.bar_type == EnumBarType.day:
                    begin_date = TradingDayHelper.get_pre_trading_day(trading_date, interval - 1)
                    bar_provider = BarProvider()
                    bar_provider.create_bar_provider_by_trading_date(self.instrument_manager, instrument_id, trading_date, interval, bar_type, *instrument_ids)
                    bar_series_day = self.load_day_bar_series(market, instrument_id, begin_date, end_trading_date)
                    trading_date = begin_date
                    for bar in bar_series_day:
                        bar_provider.add_bar(bar)
                else:
                    raise Exception("不支持的k线:" + str(bar_type))

                for bar in bar_provider.bar_series:
                    bar_series.insert(0, bar)

                if len(bar_series) >= max_length:
                    break

        bar_provider = BarProvider()
        bar_provider.create_bar_provider_by_trading_date(self.instrument_manager, instrument_id, self.trading_day, interval, bar_type, *instrument_ids)
        for bar in bar_series:
            bar_provider.bar_series.append(bar)

        # 组装当日数据
        self.__combine_today_tick(bar_provider, market, end_time)

        bar_providers.append(bar_provider)
        remove_length = len(bar_provider.bar_series) - max_length
        if remove_length > 0:
            del bar_provider.bar_series[0:remove_length]

        return bar_provider.bar_series

    def __combine_today_tick(self, bar_provider, market, end_time):
        date_time_slice = self.instrument_manager.get_living_date_time(self.trading_day, bar_provider.instrument_id)
        if end_time > date_time_slice.begin_time:
            # 加载当日数据
            tick_series = self.__load_today_tick_series(market, bar_provider.instrument_id, self.trading_day)

            if (bar_provider.bar_type == EnumBarType.second) | (bar_provider.bar_type == EnumBarType.minute) | (bar_provider.bar_type == EnumBarType.hour):
                for tick in tick_series:
                    if tick.date_time <= end_time:
                        bar_provider.add_tick(tick)
                        if tick.date_time == date_time_slice.BeginTime:  # 第1个的Tick包含集合竞价信息, 特殊处理
                            bar_provider.bar_series[-1].high = tick.high_price
                            bar_provider.bar_series[-1].open = tick.open_price
                            bar_provider.bar_series[-1].low = tick.low_price
            elif bar_provider.bar_type == EnumBarType.day:
                # // 大于1不组当日日线 和MQ一致(个人认为应该组)
                if bar_provider.interval > 1:
                    return

                if (tick_series is None) | (len(tick_series) == 0):
                    return

                first_tick = tick_series[0]
                temp_tick_series = []
                for data in tick_series:
                    if data.date_time <= end_time:
                        temp_tick_series.append(data)
                if len(temp_tick_series) == 0:
                    return
                last_tick = temp_tick_series[-1]
                if (len(bar_provider.bar_series) == 0) | (bar_provider.bar_series[-1].trading_date != first_tick.trading_day):
                    bar = Bar()
                    bar.trading_date = first_tick.trading_day
                    if len(bar_provider.bar_series) == 0:
                        temp_bar = None
                    else:
                        temp_bar = bar_provider.bar_series[-1]
                    bar.open_bar(first_tick.trading_day, first_tick, temp_bar)
                    bar.turnover = first_tick.turnover
                    bar.volume = first_tick.volume
                    bar_provider.bar_series.append(bar)

                    if last_tick is None:
                        last_tick = first_tick
                    bar_provider.bar_series[-1].high = last_tick.high_price
                    bar_provider.bar_series[-1].low = last_tick.low_price
                    bar_provider.bar_series[-1].turnover = last_tick.turnover
                    bar_provider.bar_series[-1].volume = last_tick.volume
                    bar_provider.bar_series[-1].open_interest = last_tick.open_interest
                    bar_provider.bar_series[-1].end_time = last_tick.date_time
                    bar_provider.bar_series[-1].close = last_tick.last_price
                    bar_provider.bar_series[-1].trading_date = last_tick.trading_day

    def load_last_tick(self, market, instrument_id, trading_date):
        """
        获取交易日最后一个Tick
        :param market:
        :param instrument_id:
        :param trading_date:
        :return:
        """
        path = self.__future_tick_path
        try:
            file_path = os.path.join(path, trading_date.strftime('%Y%m%d'))
            file_path = os.path.join(file_path, instrument_id.split('.')[0] + ".tk")
            if instrument_id.lower() != 'index':
                instrument = self.instrument_manager[instrument_id.split('.')[0]]
                tick_stream = BinaryTickStream(market, instrument_id, instrument.exchange_id, file_path)
                tick = tick_stream.read_last_tick()
                return tick
        except Exception as e:
            print(str(e))
            print('load_last_tick:未取到Tick数据...,请注意.....')
            return None

    def load_pre_close_by_tick(self, market, instrument_id, trading_date):
        """
        从Tick获取pre_close
        :param market:
        :param instrument_id:
        :param trading_date:
        :return:
        """
        last_tick = self.load_last_tick(market, instrument_id, trading_date)

        if last_tick is not None:
            pre_close = last_tick.pre_close_price
            if pre_close != 0:
                return pre_close
            else:
                print('pre_close:{0}为0,请注意.....'.format(pre_close))
                return pre_close
        else:
            print('load_pre_close_by_tick:未取到Tick数据,PreClose设置为0...,请注意.....')
            return 0

    def load_close_price_by_tick(self, market, instrument_id, trading_date):
        """
        从Tick获取close
        :param market:
        :param instrument_id:
        :param trading_date:
        :return:
        """
        last_tick = self.load_last_tick(market, instrument_id, trading_date)

        if last_tick is not None:
            close_price = last_tick.last_price
            if close_price != 0:
                return close_price
            else:
                print('close_price:{0}为0,请注意.....'.format(close_price))
                return close_price
        else:
            print('load_close_price_by_tick:未取到Tick数据,ClosePrice设置为0...,请注意.....')
            return 0

    def load_pre_settlement_price_by_tick(self, market, instrument_id, trading_date):
        """
        从Tick获取pre_settlement_price
        :param market:
        :param instrument_id:
        :param trading_date:
        :return:
        """
        path = self.__future_tick_path
        tick_series = []
        try:
            file_path = os.path.join(path, trading_date.strftime('%Y%m%d'))
            file_path = os.path.join(file_path, instrument_id.split('.')[0] + ".tk")
            if instrument_id.lower() != 'index':
                instrument = self.instrument_manager[instrument_id.split('.')[0]]
                tick_stream = BinaryTickStream(market, instrument_id, instrument.exchange_id, file_path)
                tick_stream.read_by_count(tick_series, 0, 5)
        except Exception as e:
            print(str(e))

        if len(tick_series) > 0:
            pre_settlement_price = tick_series[-1].pre_settlement_price
            if pre_settlement_price != 0:
                return pre_settlement_price
            else:
                print('pre_settlement_price:{0}为0,请注意.....'.format(pre_settlement_price))
                return pre_settlement_price
        else:
            print('load_pre_settlement_price_by_tick:未取到Tick数据...,请注意.....')
            return 0

    def on_bar(self, bar):
        """
        实盘on_bar维护策略获取的K线
        :param bar:
        :return:
        """
        if bar.instrument_id in self.__bar_series_map.keys():
            bar_providers = self.__bar_series_map[bar.instrument_id]
            for bar_provider in bar_providers:
                bar_provider.add_bar(bar)
        else:
            return

    def on_tick(self, tick):
        """
        实盘on_tick维护策略获取的K线
        :param tick:
        :return:
        """
        if tick.instrument_id in self.__bar_series_map.keys():
            bar_providers = self.__bar_series_map[tick.instrument_id]
            for bar_provider in bar_providers:
                bar_provider.add_tick(tick)
        else:
            return

    def load_min_bar_data_frame(self, market, instrument_id, *trading_dates, columns=None):
        """

        :param market:
        :param instrument_id:
        :param columns:
        :param trading_dates:
        :return:
        """
        bar_series = self.load_min_bar_series(market, instrument_id, *trading_dates)
        lst_bar = []
        for bar in bar_series:
            lst_bar.append([bar.trading_date, bar.begin_time, bar.end_time,
                            bar.high, bar.open, bar.low, bar.close, bar.volume, bar.turnover, bar.pre_close, bar.open_interest])

        df_bar = pd.DataFrame(lst_bar, columns=['trading_date', 'begin_time', 'end_time',
                                                'high', 'open', 'low', 'close', 'volume', 'turnover', 'pre_close', 'open_interest'])

        if columns is None:
            return df_bar
        else:
            return df_bar[columns]

    def load_day_bar_data_frame(self, market, instrument_id, *trading_dates, columns=None):
        """

        :param market:
        :param instrument_id:
        :param columns:
        :param trading_dates:
        :return:
        """
        bar_series = self.load_day_bar_series(market, instrument_id, *trading_dates)
        lst_bar = []
        for bar in bar_series:
            lst_bar.append([bar.trading_date, bar.begin_time, bar.end_time,
                            bar.high, bar.open, bar.low, bar.close, bar.volume, bar.turnover, bar.pre_close, bar.open_interest])

        df_bar = pd.DataFrame(lst_bar, columns=['trading_date', 'begin_time', 'end_time',
                                                'high', 'open', 'low', 'close', 'volume', 'turnover', 'pre_close', 'open_interest'])

        if columns is None:
            return df_bar
        else:
            return df_bar[columns]

    def load_tick_data_frame(self, market, instrument_id, *trading_dates, columns=None):
        """

        :param market:
        :param instrument_id:
        :param columns:['instrument_id', 'exchange_id', 'trading_date', 'local_time', 'date_time',
                                        'high_price', 'open_price', 'low_price', 'last_price', 'volume', 'turnover', 'open_interest',
                                        'bid_price1', 'bid_volume1', 'ask_price1', 'ask_volume1', 'up_limit', 'drop_limit',
                                        'pre_close_price', 'pre_settlement_price', 'pre_open_interest']
        :param trading_dates:
        :return:
        """
        tick_series = self.load_tick_series(market, instrument_id, *trading_dates)
        lst_tick = []
        for tick in tick_series:
            lst_tick.append([tick.instrument_id, tick.exchange_id, tick.trading_date, tick.local_time, tick.date_time,
                             tick.high_price, tick.open_price, tick.low_price, tick.last_price, tick.volume, tick.turnover, tick.open_interest,
                             tick.bid_price1, tick.bid_volume1, tick.ask_price1, tick.ask_volume1, tick.up_limit, tick.drop_limit,
                             tick.pre_close_price, tick.pre_settlement_price, tick.pre_open_interest])

        df_tick = pd.DataFrame(lst_tick,
                               columns=['instrument_id', 'exchange_id', 'trading_date', 'local_time', 'date_time',
                                        'high_price', 'open_price', 'low_price', 'last_price', 'volume', 'turnover', 'open_interest',
                                        'bid_price1', 'bid_volume1', 'ask_price1', 'ask_volume1', 'up_limit', 'drop_limit',
                                        'pre_close_price', 'pre_settlement_price', 'pre_open_interest'])

        if columns is None:
            return df_tick
        else:
            return df_tick[columns]

    def load_tick_data_frame_by_date_time(self, market, instrument_id, begin_time, end_time, columns=None):
        """
        根据时间区间加载历史Tick数据
        :param columns:
        :param market:市场
        :param instrument_id:合约编号
        :param begin_time:开始时间
        :param end_time:截止时间
        :return:
        """
        tick_series = self.load_tick_series_by_date_time(market, instrument_id, begin_time, end_time)
        lst_tick = []
        for tick in tick_series:
            lst_tick.append([tick.instrument_id, tick.exchange_id, tick.trading_date, tick.local_time, tick.date_time,
                             tick.high_price, tick.open_price, tick.low_price, tick.last_price, tick.volume, tick.turnover, tick.open_interest,
                             tick.bid_price1, tick.bid_volume1, tick.ask_price1, tick.ask_volume1, tick.up_limit, tick.drop_limit,
                             tick.pre_close_price, tick.pre_settlement_price, tick.pre_open_interest])

        df_tick = pd.DataFrame(lst_tick,
                               columns=['instrument_id', 'exchange_id', 'trading_date', 'local_time', 'date_time',
                                        'high_price', 'open_price', 'low_price', 'last_price', 'volume', 'turnover', 'open_interest',
                                        'bid_price1', 'bid_volume1', 'ask_price1', 'ask_volume1', 'up_limit', 'drop_limit',
                                        'pre_close_price', 'pre_settlement_price', 'pre_open_interest'])

        if columns is None:
            return df_tick
        else:
            return df_tick[columns]

    def load_bar_data_frame_by_date(self, market, instrument_id, interval, bar_type, begin_date, end_date, *instrument_ids, columns=None):
        """
        返回一个DataFrame
        :param columns:None为全选，可选列表['trading_date', 'begin_time', 'end_time', 'high', 'open',  'low', 'close', 'volume', 'turnover', 'pre_close', 'open_interest']
        :param market:
        :param instrument_id:
        :param interval:
        :param bar_type:
        :param begin_date:
        :param end_date:
        :param instrument_ids:
        :return:
        """
        bar_series = self.load_bar_series_by_date(market, instrument_id, interval, bar_type, begin_date, end_date, *instrument_ids)
        lst_bar = []
        for bar in bar_series:
            lst_bar.append([bar.trading_date, bar.begin_time, bar.end_time,
                            bar.high, bar.open, bar.low, bar.close, bar.volume, bar.turnover, bar.pre_close, bar.open_interest])

        df_bar = pd.DataFrame(lst_bar, columns=['trading_date', 'begin_time', 'end_time',
                                                'high', 'open', 'low', 'close', 'volume', 'turnover', 'pre_close', 'open_interest'])

        if columns is None:
            return df_bar
        else:
            return df_bar[columns]

    def load_bar_data_frame_by_date_time(self, market, instrument_id, interval, bar_type, begin_time, end_time, *instrument_ids, columns=None):
        """

        :param market:
        :param instrument_id:
        :param interval:
        :param bar_type:
        :param begin_time:
        :param end_time:
        :param columns:
        :param instrument_ids:
        :return:
        """
        bar_series = self.load_bar_series_by_date_time(market, instrument_id, interval, bar_type, begin_time, end_time, *instrument_ids)
        lst_bar = []
        for bar in bar_series:
            lst_bar.append([bar.trading_date, bar.begin_time, bar.end_time,
                            bar.high, bar.open, bar.low, bar.close, bar.volume, bar.turnover, bar.pre_close, bar.open_interest])

        df_bar = pd.DataFrame(lst_bar, columns=['trading_date', 'begin_time', 'end_time',
                                                'high', 'open', 'low', 'close', 'volume', 'turnover', 'pre_close', 'open_interest'])

        if columns is None:
            return df_bar
        else:
            return df_bar[columns]

    def load_bar_data_frame_by_length(self, market, instrument_id, interval, bar_type, max_length, end_date, *instrument_ids, columns=None):
        """

        :param market:
        :param instrument_id:
        :param interval:
        :param bar_type:
        :param max_length:
        :param end_date:
        :param columns:
        :param instrument_ids:
        :return:
        """
        bar_series = self.load_bar_series_by_length(market, instrument_id, interval, bar_type, max_length, end_date, *instrument_ids)
        lst_bar = []
        for bar in bar_series:
            lst_bar.append([bar.trading_date, bar.begin_time, bar.end_time,
                            bar.high, bar.open, bar.low, bar.close, bar.volume, bar.turnover, bar.pre_close, bar.open_interest])

        df_bar = pd.DataFrame(lst_bar, columns=['trading_date', 'begin_time', 'end_time',
                                                'high', 'open', 'low', 'close', 'volume', 'turnover', 'pre_close', 'open_interest'])

        if columns is None:
            return df_bar
        else:
            return df_bar[columns]



from SapphireQuant.Core import Exchange


class FutureExchange(Exchange):
    """
    期货交易所
    """
    def __init__(self):
        super().__init__()
        self.products = []
        self.product_map = {}


import datetime
import uuid

from SapphireQuant.Core.Enum.EnumOrderSubmitStatus import EnumOrderSubmitStatus

from SapphireQuant.Core.Enum.EnumOrderStatus import EnumOrderStatus

from SapphireQuant.Core.Entity import Entity
from SapphireQuant.Core.Enum import EnumBuySell, EnumMarket, EnumOpenClose, EnumHedgeFlag


class FutureOrder(Entity):
    """
    期货委托
    """

    def __init__(self):
        super().__init__()
        self.task_guid = uuid.uuid1()
        self.local_id = ''
        self.broker_id = ''
        self.investor_id = ''
        self.user_id = ''
        self.order_sys_id = ''
        self.instrument_id = ''
        self.instrument_name = ''
        self.instrument_type = EnumMarket.期货
        self.exchange_id = ''
        self.direction = EnumBuySell.买入
        self.open_close = EnumOpenClose.开仓
        self.hedge_flag = EnumHedgeFlag.投机
        self.order_status = EnumOrderStatus.未知
        self.order_submit_status = EnumOrderSubmitStatus.un_known
        self.price_type = 0
        self.order_price = 0.0
        self.order_volume = 0
        self.order_time = datetime.datetime.today()
        self.volume_traded = 0
        self.volume_left = self.order_volume - self.volume_traded
        self.volume_canceled = 0
        self.cancel_time = datetime.datetime.today()
        self.order_reject_reason = 0
        self.match_price = 0
        self.match_amount = 0
        self.frozen_amount = 0
        self.transaction_cost = 0
        self.order_type = 0
        self.local_order_time = datetime.datetime.today()
        self.status_msg = ''
        self.remark = ''
        self.is_completed = True

    def update(self, order):
        """
        更新
        :param order:
        """
        if self.is_new(order):
            self.order_sys_id = order.order_sys_id
            self.order_status = order.order_status
            self.order_price = order.order_price
            self.volume_traded = order.volume_traded
            self.match_amount = order.match_amount
            self.order_time = order.order_time
            self.local_order_time = order.local_order_time
            self.cancel_time = order.cancel_time
            self.remark = order.remark
            self.order_reject_reason = order.order_reject_reason

    def is_new(self, new_order):
        """

        :param new_order:
        :return:
        """
        if self.volume_traded <= new_order.volume_traded:
            return True

        if self.volume_traded > new_order.volume_traded:
            return False

        if self.is_completed:
            return False

        return True

    def to_string(self):
        """

        :return:
        """
        return '{0},状态[{1}],'.format(str(self.guid), self.order_status.chinese_name()) + '委托序号:{0},委托时间:{1},'.format(
            self.order_sys_id, self.order_time.strftime('%Y/%m/%d %H:%M:%S')) + '{0}@{1},[{2}][{3}]'.format(
            self.instrument_type.name, self.exchange_id, self.product_code, self.strategy_id) + '{0}{1}{2}{3}手[{4}]'.format(
            self.hedge_flag.chinese_name(), self.direction.chinese_name(), self.open_close.chinese_name(), self.order_volume, self.instrument_id) + '{0},剩:{1},'.format(
            self.order_price, self.order_volume - self.volume_traded) + '成交量:{0},成交金额:{1},撤:{2},'.format(
            self.volume_traded, self.match_amount, self.volume_canceled) + '{0},{1}'.format(self.order_reject_reason, self.status_msg)



import datetime

from SapphireQuant.Config.TradingDay.TradingDayHelper import TradingDayHelper


class YfTimeHelper:
    """
    夜盘帮助
    """
    @staticmethod
    def join_date_time(trading_day, pre_trading_day1, pre_trading_day2, time_span):
        """

        :param trading_day:
        :param pre_trading_day1:
        :param pre_trading_day2:
        :param time_span:
        :return:
        """
        temp = datetime.datetime(2000, 1, 1) + time_span
        if temp.hour < 7:
            return pre_trading_day2 + time_span
        if temp.hour < 18:
            return trading_day + time_span

        return pre_trading_day1 + time_span

    @staticmethod
    def get_trading_day(date_time):
        """
        获取交易日
        :param date_time:datetime.datetime
        :return:datetime.datetime
        """
        trading_date = datetime.datetime(date_time.year, date_time.month, date_time.day)
        if date_time.hour <= 6:
            trading_date = TradingDayHelper.get_first_trading_day(trading_date)
        elif date_time.hour >= 16:
            trading_date = TradingDayHelper.get_next_trading_day(trading_date)
        return trading_date

    @staticmethod
    def get_date_time_by_trading_day(trading_day, time_span):
        """

        :param trading_day:
        :param time_span:
        :return:
        """
        temp = datetime.datetime(2000, 1, 1) + time_span
        if temp.hour < 7:
            if trading_day.weekday() != 0:
                return trading_day + time_span

            return trading_day + datetime.timedelta(days=-2) + time_span

        if temp.hour < 18:
            return trading_day + time_span

        if trading_day.weekday() != 0:
            return trading_day + datetime.timedelta(days=-1) + time_span

        return trading_day + datetime.timedelta(days=-3) + time_span



"""
Event-driven framework of sapphire-quant framework.
"""
from typing import Any, Callable


class Event:
    """
    Event object consists of a type string which is used
    by event engine for distributing event, and a data
    object which contains the real data.
    """

    def __init__(self, type: str, data: Any = None):
        """"""
        self.type: str = type
        self.data: Any = data


# Defines handler function to be used in event engine.
HandlerType = Callable[[Event], None]



import datetime
import uuid

from SapphireQuant.Core.Entity import Entity
from SapphireQuant.Core.Enum import EnumMarket, EnumBuySell, EnumOpenClose, EnumHedgeFlag


class FutureTrade(Entity):
    """
    期货成交
    """
    def __init__(self):
        super().__init__()
        self.task_guid = uuid.uuid1()
        self.local_id = ''
        self.broker_id = ''
        self.investor_id = ''
        self.user_id = ''
        self.order_guid = ''
        self.order_sys_id = ''
        self.trade_sys_id = ''
        self.instrument_id = ''
        self.instrument_name = ''
        self.market = EnumMarket.期货
        self.exchange_id = ''
        self.direction = EnumBuySell.买入
        self.open_close = EnumOpenClose.开仓
        self.hedge_flag = EnumHedgeFlag.套保
        self.trade_volume = 0
        self.trade_price = 0
        self.trade_amount = 0
        self.trade_time = datetime.datetime.today()
        self.remark = ''

    def to_string(self):
        """

        :return:
        """
        return '期货成交:{0}{1}{2}手[{3}]成交价{4}@{5},{6}'.format(
            self.direction.chinese_name(), self.open_close.chinese_name(), self.trade_volume, self.instrument_id, self.trade_price, self.trade_time.strftime('%H:%M:%S.%f'), self.remark)



import datetime
import os
import sys
from decimal import Decimal

from SapphireQuant.Core.Bar import Bar
from SapphireQuant.DataProcessing.BinaryReader import BinaryReader
from SapphireQuant.DataProcessing.BinaryWriter import BinaryWriter


class BinaryDayBarStream(object):
    """
    日线读取流
    """
    CFileHeaderLen = 64
    CBarLen = 12 * 4
    CReserve = 48  # 空的预留字段
    CPosEnd = 8

    def __init__(self, market, instrument_id, file_path):
        self._tick_count = 0
        self._valid = False
        self._first_read = False
        self._begin_date = datetime.datetime.max
        self._end_date = datetime.datetime.max

        self._instrument_id = instrument_id.split('.')[0]
        self._file_path = file_path
        self._first_read = True

    def _read_header(self, reader):
        """
        读取头文件
        :param reader:
        """
        reader.stream.seek(0, 0)
        interval = reader.read_int16()
        bar_type = reader.read_int16()
        if (bar_type != 4) or (interval != 1):
            print("Wrong Day Kline Flag!!!")

        year = reader.read_int16()
        month = reader.read_byte()
        day = reader.read_byte()
        self._begin_date = datetime.datetime(year, month, day)

        year = reader.read_int16()
        month = reader.read_byte()
        day = reader.read_byte()
        self._end_date = datetime.datetime(year, month, day)

        if (self._begin_date == datetime.datetime.max) or (self._end_date == datetime.datetime.max) or (self._begin_date > self._end_date):
            raise Exception("Invalid Kline Date")

        self._tick_count = reader.read_int32()
        length = os.path.getsize(self._file_path)
        if self._tick_count != int(((length - self.CFileHeaderLen) / self.CBarLen)):
            raise Exception("Day KLine File Broken!!!")
        self._first_read = False

    def _read_bars(self, reader, bar_series, begin_date, end_date):
        """
        读取bars
        :param reader:
        :param bar_series:
        :param begin_date:
        :param end_date:
        """
        reader.stream.seek(self.CFileHeaderLen, 0)
        begin_date = datetime.datetime(begin_date.year, begin_date.month, begin_date.day)
        end_date = datetime.datetime(end_date.year, end_date.month, end_date.day)
        i = 0
        for i in range(0, self._tick_count):
            year = reader.read_int16()
            month = reader.read_byte()
            day = reader.read_byte()
            reader.stream.seek(self.CBarLen - 4, 1)
            trading_date = datetime.datetime(year, month, day)
            if trading_date >= begin_date:
                reader.stream.seek(-self.CBarLen, 1)
                break

        for index in range(i, self._tick_count):
            year = reader.read_int16()
            month = reader.read_byte()
            day = reader.read_byte()

            trading_date = datetime.datetime(year, month, day)

            if trading_date > end_date:
                break

            bar = Bar()
            bar.begin_time = trading_date
            bar.end_time = trading_date
            bar.open = reader.read_int32() / 1000.0
            bar.close = reader.read_int32() / 1000.0
            bar.high = reader.read_int32() / 1000.0
            bar.low = reader.read_int32() / 1000.0
            bar.pre_close = reader.read_int32() / 1000.0
            bar.volume = reader.read_double()
            bar.turnover = reader.read_double()
            bar.open_interest = reader.read_double()
            # bar.IsCompleted = True
            bar.trading_date = trading_date
            bar.instrument_id = self._instrument_id

            bar_series.append(bar)

    def read(self, bar_series, begin_date, end_date):
        """
        读取日线
        :param bar_series:
        :param begin_date:
        :param end_date:
        :return:
        """
        try:
            if not os.path.exists(self._file_path):
                self._first_read = True
                self._valid = False
                print("Read Day Kline Failed,Do Not Exist File:{0}".format(self._file_path))
                return False

            stream = open(self._file_path, 'rb+')
            length = os.path.getsize(self._file_path)
            if (length - self.CFileHeaderLen) % self.CBarLen != 0:
                raise Exception("Day Kline File Is Broken,Data Wrong!!!")

            reader = BinaryReader(stream)

            if self._first_read:
                self._read_header(reader)

            self._read_bars(reader, bar_series, begin_date, end_date)
            self._valid = True
            return True
        except Exception as e:
            self._first_read = True
            self._valid = False
            print("Read Day Kline Failed:{0},{1}".format(self._file_path, str(e)))
        finally:
            pass

        return False

    def read_all(self, bar_series):
        """

        :param bar_series:
        :return:
        """
        try:
            if not os.path.exists(self._file_path):
                self._first_read = True
                self._valid = False
                print("Read Day Kline Failed,Do Not Exist File:{0}".format(self._file_path))
                return False

            stream = open(self._file_path, 'rb+')
            length = os.path.getsize(self._file_path)
            if (length - self.CFileHeaderLen) % self.CBarLen != 0:
                raise Exception("Day Kline File Is Broken,Data Wrong!!!")

            reader = BinaryReader(stream)
            if self._first_read:
                self._read_header(reader)

            self._read_bars(reader, bar_series, self._begin_date, self._end_date)
            self._valid = True
            return True
        except Exception as e:
            self._first_read = True
            self._valid = False
            print("Read Day Kline Failed:{0},{1}".format(self._file_path, str(e)))
        finally:
            pass

        return False

    def _write_header(self, writer):
        writer.stream.seek(0, 0)
        writer.write_int16(1)
        writer.write_int16(4)
        writer.write_int16(self._begin_date.year)
        writer.write_byte(self._begin_date.month)
        writer.write_byte(self._begin_date.day)
        writer.write_int16(self._end_date.year)
        writer.write_byte(self._end_date.month)
        writer.write_byte(self._end_date.day)
        writer.write_int32(self._tick_count)

        buffer = bytearray(self.CReserve)
        writer.write_bytes(buffer)

    def _write_bars(self, bar_series, writer, start, count):
        end = start+count
        for i in range(start, end, 1):
            bar = bar_series[i]

            writer.write_int16(bar.begin_time.year)
            writer.write_byte(bar.begin_time.month)
            writer.write_byte(bar.begin_time.day)

            self.write_float_with_decimal(writer, bar.open)
            self.write_float_with_decimal(writer, bar.close)
            self.write_float_with_decimal(writer, bar.high)
            self.write_float_with_decimal(writer, bar.low)
            self.write_float_with_decimal(writer, bar.pre_close)
            writer.write_double(bar.volume)
            writer.write_double(bar.turnover)
            writer.write_double(bar.open_interest)

    @staticmethod
    def write_float_with_decimal(writer, value):
        """

        :param writer:
        :param value:
        """
        if value < sys.float_info.max:
            value = Decimal.from_float(value)
            writer.write_int32(int(round(value * Decimal.from_float(1000), 0)))
        else:
            writer.write_int32(0)

    def write(self, bar_series):
        """

        :param bar_series:
        """
        if bar_series is None or len(bar_series) == 0:
            return True

        try:
            with open(self._file_path, 'wb') as stream:
                self._tick_count = len(bar_series)
                self._begin_date = bar_series[0].trading_date
                self._end_date = bar_series[-1].trading_date

                writer = BinaryWriter(stream)
                self._write_header(writer)
                self._write_bars(bar_series, writer, 0, self._tick_count)

            return True
        except Exception as e:
            print('Write Day Bar Error:{0}'.format(str(e)))
        return False

    def append(self, bar_series):
        """
        Append Bar Series
        :param bar_series:
        """
        if not os.path.exists(self._file_path) or sys.getsizeof(self._file_path) < self.CFileHeaderLen:
            return self.write(bar_series)

        try:
            with open(self._file_path, 'rb+') as stream:
                reader = BinaryReader(stream)
                self._read_header(reader)
                count = len(bar_series)

                self._tick_count = int((sys.getsizeof(self._file_path) - self.CFileHeaderLen)/self.CBarLen)
                self._tick_count += count
                self._end_date = bar_series[-1].trading_date

                writer = BinaryWriter(stream)
                self._write_header(writer)
                writer.stream.seek(0, 2)
                self._write_bars(bar_series, writer, 0, count)

            return True
        except Exception as e:
            print('Append Day Bar Error:{0}'.format(str(e)))
        return False



import datetime
import os
import re
import xml

import xml.dom.minidom


from SapphireQuant.Config.TradingFrame.FutureProductTradingFrame import FutureProductTradingFrame
from SapphireQuant.Config.TradingFrame.FutureExchangeTradingFrame import FutureExchangeTradingFrame
from SapphireQuant.Config.TradingFrame.FutureTradingFrame import FutureTradingFrame
from SapphireQuant.Config.TradingFrame.LimitValue import LimitValue
from SapphireQuant.Config.TradingFrame.TradingFrameNode import TradingFrameNode
from SapphireQuant.Core.Enum.EnumMarket import EnumMarket
from SapphireQuant.Core.TimeSlice import TimeSlice


class TradingFrameManager:
    """
    交易时间框架管理
    """
    FileName = "TradingFrame.xml"

    def __init__(self):
        self.__is_loaded = False
        self.__path = ""
        self.__future = None
        self.__future_trading_time_map = {}
        self.__default_frame = []
        self.__default_future_trading_time_map = {}

    @property
    def future(self):
        """
        期货交易时间
        :return:
        """
        return self.__future

    def load(self, config_directory):
        """
        加载配置
        :param config_directory:
        """
        if not self.__is_loaded:
            try:
                self.__path = os.path.join(config_directory, self.FileName)
                if os.path.exists(self.__path):
                    pass
                else:
                    print("未找到配置文件:" + self.__path)

                _root = xml.dom.minidom.parse(self.__path).documentElement
                future_nodes = _root.getElementsByTagName('Future')
                for future_node in future_nodes:
                    self.__future = TradingFrameManager.__read_future_trading_frame(future_node)

                self.__default_frame = []
                time_slice_am = TimeSlice()
                time_slice_am.begin_time = datetime.timedelta(hours=9, minutes=30, seconds=0)
                time_slice_am.end_time = datetime.timedelta(hours=11, minutes=30, seconds=0)
                self.__default_frame.append(time_slice_am)

                time_slice_pm = TimeSlice()
                time_slice_pm.begin_time = datetime.timedelta(hours=13, minutes=0, seconds=0)
                time_slice_pm.end_time = datetime.timedelta(hours=15, minutes=0, seconds=0)
                self.__default_frame.append(time_slice_pm)

                self.__default_future_trading_time_map = {}
                self.__future_trading_time_map = {}

                for exchange in self.__future.exchanges:
                    self.__default_future_trading_time_map[exchange.id] = exchange.trading_time_slices

                    for product in exchange.products:
                        self.__future_trading_time_map[product.id] = product.trading_frames

                self.__is_loaded = True
            except Exception as e:
                print(str(e))
        pass

    def get_trading_time_slices(self, trading_date, market, exchange_id, product_id=""):
        """
        获取指定交易日的交易时间区间TimeSlice
        :param trading_date:
        :param market:
        :param exchange_id:
        :param product_id:
        :return:
        """
        if trading_date is None:
            return self.get_last_trading_time_slices(market, exchange_id, product_id)

        trading_frame_slices = self.get_trading_time_frames(market, exchange_id, product_id)

        lst_time_slices = None
        i = len(trading_frame_slices) - 1
        while i >= 0:
            trading_frame = trading_frame_slices[i]
            begin_day = trading_frame.begin_day

            if trading_date >= begin_day:
                lst_time_slices = trading_frame.trading_time_slices
                break

            i -= 1

        if lst_time_slices is not None:
            return lst_time_slices

        return self.get_default_trading_time_slices(market, exchange_id)

    def get_default_trading_time_slices(self, market, exchange_id):
        """
        获取默认的TimeSlice
        :param market:
        :param exchange_id:
        :return:
        """
        if market == EnumMarket.期货:
            if exchange_id in self.__default_future_trading_time_map.keys():
                return self.__default_future_trading_time_map[exchange_id]

    def get_last_trading_time_slices(self, market, exchange_id, product_id=""):
        """
        获取最后一个TimeSlice
        :param market:
        :param exchange_id:
        :param product_id:
        :return:
        """
        time_frame_series = self.get_trading_time_frames(market, exchange_id, product_id)

        if time_frame_series is not None:
            return time_frame_series[-1].trading_time_slices

    def get_trading_time_frames(self, market, exchange_id, product_id=""):
        """
        获取交易时间
        :param market:
        :param exchange_id:
        :param product_id:
        :return:
        """
        if market == EnumMarket.期货:
            if product_id in self.__future_trading_time_map.keys():
                return self.__future_trading_time_map[product_id]
        return None

    def get_living_time(self, trading_day, market, exchange_id, product_id=""):
        """
        根据交易日 获取当日的开盘收盘时间的TimeSlice
        :param trading_day:
        :param market:
        :param exchange_id:
        :param product_id:
        :return:
        """
        lst_time_slices = self.get_trading_time_slices(trading_day, market, exchange_id, product_id)

        if (lst_time_slices is not None) & (len(lst_time_slices) > 0):
            time_slices = TimeSlice()
            time_slices.begin_time = lst_time_slices[0].begin_time
            time_slices.end_time = lst_time_slices[-1].end_time
            return time_slices

        return None

    def get_open_time(self, trading_day, market, exchange_id, product_id=""):
        """
        获取当日开盘时间
        :param trading_day:
        :param market:
        :param exchange_id:
        :param product_id:
        :return:
        """
        lst_time_slices = self.get_trading_time_slices(trading_day, market, exchange_id, product_id)

        if (lst_time_slices is not None) & (len(lst_time_slices) > 0):
            return lst_time_slices[0].begin_time

        return None

    def get_close_time(self, trading_day, market, exchange_id, product_id=""):
        """
        获取当日收盘时间
        :param trading_day:
        :param market:
        :param exchange_id:
        :param product_id:
        :return:
        """
        lst_time_slices = self.get_trading_time_slices(trading_day, market, exchange_id, product_id)

        if (lst_time_slices is not None) & (len(lst_time_slices) > 0):
            return lst_time_slices[-1].end_time

        return None

    def get_listing_date(self, market, exchange_id, instrument_id):
        """
        获取上市日期
        """
        product_id = re.sub(r"\d", "", instrument_id, 0)
        lst_data = self.get_trading_time_frames(market, exchange_id, product_id)
        if lst_data is None:
            return None
        lst_data.sort(key=lambda x: x.begin_day)
        if len(lst_data) >= 1:
            return lst_data[1].begin_day
        else:
            return None

    def get_active_date(self, market, exchange_id, instrument_id):
        """
        获取活跃日期
        """
        product_id = re.sub(r"\d", "", instrument_id, 0)
        lst_data = self.get_trading_time_frames(market, exchange_id, product_id)
        if lst_data is None:
            return None
        lst_data.sort(key=lambda x: x.begin_day)
        if product_id == 'fu':
            return datetime.datetime(2018, 7, 17)
        else:
            if len(lst_data) >= 1:
                return lst_data[1].begin_day
            else:
                return None

    def get_product_ids(self, market, exchange_id):
        """
        获取产品ID
        :param market:
        :param exchange_id:
        :return:
        """
        lst_product_id = []
        if market == EnumMarket.期货:
            for exchange in self.__future.exchanges:
                if exchange.id == exchange_id:
                    for product in exchange.products:
                        lst_product_id.append(product.id)

        return lst_product_id

    def get_all_product_ids(self, market):
        """
        获取产品ID
        :param market:
        :return:
        """
        lst_product_id = []
        if market == EnumMarket.期货:
            for exchange in self.__future.exchanges:
                for product in exchange.products:
                    lst_product_id.append(product.id)

        return lst_product_id

    @staticmethod
    def __read_future_trading_frame(node):
        future_trading_frame = FutureTradingFrame()
        future_trading_frame.exchanges = []
        exchange_nodes = node.getElementsByTagName('Exchange')
        for exchange_node in exchange_nodes:
            exchange_trading_frame = TradingFrameManager.__read_future_exchange_trading_frame(exchange_node)
            future_trading_frame.exchanges.append(exchange_trading_frame)
        return future_trading_frame

    @staticmethod
    def __read_future_exchange_trading_frame(node):
        future_exchange_trading_frame = FutureExchangeTradingFrame()
        future_exchange_trading_frame.id = node.getAttribute('id')
        future_exchange_trading_frame.name = node.getAttribute('name')
        future_exchange_trading_frame.trading_time = node.getAttribute('TradingTime')
        future_exchange_trading_frame.products = []
        products_nodes = node.getElementsByTagName('Products')
        for products_node in products_nodes:
            product_nodes = products_node.getElementsByTagName('Product')
            for product_node in product_nodes:
                product_trading_frame = TradingFrameManager.__read_product_trading_frame(product_node)
                future_exchange_trading_frame.products.append(product_trading_frame)
        return future_exchange_trading_frame

    @staticmethod
    def __read_product_trading_frame(node):
        product_trading_frame = FutureProductTradingFrame()
        product_trading_frame.id = node.getAttribute('id')
        product_trading_frame.name = node.getAttribute('name')
        trading_frame_nodes = node.getElementsByTagName('TradingFrame')
        product_trading_frame.trading_frames = []
        for trading_frame_node in trading_frame_nodes:
            trading_frame = TradingFrameNode()
            trading_frame.begin_day_string = trading_frame_node.getAttribute('BeginDay')
            trading_frame.trading_time = trading_frame_node.getAttribute('TradingTime')
            product_trading_frame.trading_frames.append(trading_frame)

        limit_value_nodes = node.getElementsByTagName('LimitValue')
        for limit_value_node in limit_value_nodes:
            product_trading_frame.limit_value = LimitValue()
            product_trading_frame.limit_value.up_limit = float(limit_value_node.getAttribute('UpLimit'))
            product_trading_frame.limit_value.low_limit = float(limit_value_node.getAttribute('LowLimit'))

        return product_trading_frame


# trading_frame_manager = TradingFrameManager()
# config_dir = os.path.join(sys.modules['ROOT_DIR'], 'Profiles')
# trading_frame_manager.load(config_dir)
# index = 0
# exchange_index_map = {}
# exchange_product_id_map = {}
# for exchange in trading_frame_manager.future.exchanges:
#     print(exchange.id + ":" + exchange.name + ":" + exchange.trading_time)
#     exchange_index = 0
#     exchange_products = ""
#     for product in exchange.products:
#         print(product.id + ":" + product.name)
#         for trading_frame in product.trading_frames:
#             print("TradingFrameNode:BeginDay=" + trading_frame.begin_day_string + ",TradingTime=" + trading_frame.trading_time)
#         print("LimitValue:" + str(product.limit_value.up_limit) + "," + str(product.limit_value.low_limit))
#         index += 1
#         exchange_index += 1
#         exchange_products += product.id + ";"
#
#     exchange_index_map[exchange.id] = exchange_index
#     exchange_product_id_map[exchange.id] = exchange_products
# print("共计" + str(index) + "个品种")
# for data in exchange_index_map.keys():
#     print(data + ":" + str(exchange_index_map[data]) + "个品种:" + exchange_product_id_map[data])
#
# market = EnumMarket.期货
# exchange_id = "CFFEX"
# product_id = "IF"
# trading_date = datetime.datetime(2019, 10, 8)
# lst_time_slices = trading_frame_manager.get_trading_time_slices(trading_date, market, exchange_id, product_id)
# print("GetTradingTimeSlices:"+exchange_id+","+product_id+","+trading_date.strftime("%Y/%m%d"))
# for time_slice in lst_time_slices:
#     print("BeginTime:"+time_slice.to_string())
#     print("EndTime:"+time_slice.to_string())
#
# default_trading_time_slices = trading_frame_manager.get_default_trading_time_slices(market, exchange_id)
# print("GetDefaultTradingTimeSlices:"+exchange_id)
# for time_slice in default_trading_time_slices:
#     print("BeginTime:"+time_slice.to_string())
#     print("EndTime:"+time_slice.to_string())
#
# last_trading_time_slices = trading_frame_manager.get_last_trading_time_slices(market, exchange_id, product_id)
# print("GetLastTradingTimeSlices:"+exchange_id)
# for time_slice in last_trading_time_slices:
#     print("BeginTime:"+time_slice.to_string())
#     print("EndTime:"+time_slice.to_string())
#
# trading_time_frames = trading_frame_manager.get_trading_time_frames(market, exchange_id, product_id)
# print("GetTradingTimeFrames:"+exchange_id+","+product_id)
# for trading_frame in trading_time_frames:
#     print("BeginDay:"+trading_frame.begin_day.strftime("%Y/%m/%d"))
#     print("TradingTime:"+trading_frame.trading_time)
#
# living_time = trading_frame_manager.get_living_time(trading_date, market, exchange_id, product_id)
# print("GetLivingTime:"+exchange_id+","+product_id+","+trading_date.strftime("%Y/%m/%d"))
# print("TradingTime:"+living_time.to_string())
#
# open_time = trading_frame_manager.get_open_time(trading_date, market, exchange_id, product_id)
# print("GetOpenTime:"+exchange_id+","+product_id+","+trading_date.strftime("%Y/%m/%d"))
# print("TradingTime:"+str(open_time))
#
# close_time = trading_frame_manager.get_close_time(trading_date, market, exchange_id, product_id)
# print("GetCloseTime:"+exchange_id+","+product_id+","+trading_date.strftime("%Y/%m/%d"))
# print("TradingTime:"+str(close_time))


from SapphireQuant.Config.Ctp.CtpCounterGroup import CtpCounterGroup


class CtpBroker:
    """
    期货Broker
    """
    def __init__(self, broker_id, name, default_group_name):
        self.id = broker_id
        self.name = name
        self.default_group_name = default_group_name
        self.ctp_counter_groups = []

    def add(self, server):
        """
        添加服务器
        :param server:
        """
        ctp_counter_group = None
        for data in self.ctp_counter_groups:
            if data.name == server.group_name:
                ctp_counter_group = data
                break
        if ctp_counter_group is None:
            ctp_counter_group = CtpCounterGroup(server.group_name)
            self.ctp_counter_groups.append(ctp_counter_group)
        ctp_counter_group.add(server)

    def get_servers_by_network_server_type(self, network_type, server_type):
        """
        获取服务器列表
        :param network_type:
        :param server_type:
        :return:
        """
        return self.get_servers(self.default_group_name, network_type, server_type)

    def get_servers(self, group_name, network_type, server_type):
        """
        获取服务器列表
        :param group_name:
        :param network_type:
        :param server_type:
        :return:
        """
        ctp_counter_group = None
        for data in self.ctp_counter_groups:
            if data.name == group_name:
                ctp_counter_group = data
                break
        if ctp_counter_group is not None:
            return ctp_counter_group.get_servers_by_network_server_type(network_type, server_type)

        print('未找到组名:{0}的服务器列表'.format(group_name))
        return None





import datetime

from SapphireQuant.Config.TradingDay import HolidayHelper
from SapphireQuant.Config.TradingDay.TradingDayHelper import TradingDayHelper
from SapphireQuant.Core.DateTimeSlice import DateTimeSlice
from SapphireQuant.Core.Enum import EnumMarket
from SapphireQuant.Core.Enum.EnumBarType import EnumBarType
from SapphireQuant.Core.TimeSlice import TimeSlice
from SapphireQuant.DataProcessing.BarProviderInfo import BarProviderInfo
from SapphireQuant.Core.YfTimeHelper import YfTimeHelper
import pandas as pd


class BaseBarHelper:
    """
    基础K线帮助类
    """
    one_day = datetime.timedelta(days=1)

    def __init__(self):
        self.__templates = []

    @staticmethod
    def create_bar_provider_info(lst_time_slice, offset, interval, bar_type):
        """

        :param lst_time_slice:
        :param offset:
        :param interval:
        :param bar_type:
        :return:
        """
        bars = BaseBarHelper.create_time_slices(lst_time_slice, offset, interval, bar_type)
        if bars is None:
            return None

        bar_provider_info = BarProviderInfo()
        bar_provider_info.time_slices = bars
        bar_provider_info.bar_type = bar_type
        bar_provider_info.interval = interval
        bar_provider_info.offset = offset
        bar_provider_info.living_time = TimeSlice()
        bar_provider_info.living_time.begin_time = lst_time_slice[0].begin_time
        bar_provider_info.living_time.end_time = lst_time_slice[-1].end_time

    @staticmethod
    def create_time_slice(lst_time_slice, offset, interval, bar_type):
        """
        切分时间片段
        根据一天的交易时间片段，按照给定的K线参数，进行时间片段的划分
        :param lst_time_slice:源生时间片段
        :param offset:位移
        :param interval:间隔
        :param bar_type:K线类型
        :return:
        """
        if len(lst_time_slice) == 0:
            return None

        offset_time_span = datetime.timedelta(0, 0, offset)
        if bar_type == EnumBarType.second:
            interval_time_span = datetime.timedelta(seconds=interval)
        elif bar_type == EnumBarType.minute:
            interval_time_span = datetime.timedelta(minutes=interval)
        elif bar_type == EnumBarType.hour:
            interval_time_span = datetime.timedelta(hours=interval)
        elif bar_type == EnumBarType.day:
            day_time_slice = TimeSlice()
            day_time_slice.begin_time = lst_time_slice[0].begin_time
            day_time_slice.end_time = lst_time_slice[-1].end_time
            day_bars = [day_time_slice]
            return day_bars
        else:
            raise Exception("不支持的K线:" + str(bar_type))

        begin_time = lst_time_slice[0].begin_time
        bars = []
        if offset > 0:
            slice = TimeSlice()
            slice.begin_time = begin_time
            slice.end_time = begin_time + offset_time_span
            if slice.end_time.days() > 0:
                slice.end_time = slice.end_time = BaseBarHelper.one_day

            bars.append(slice)

            begin_time = slice.end_time

        end_time = begin_time
        diff_time_span = datetime.timedelta(0, 0, 0)
        for i in range(len(lst_time_slice)):
            ts = lst_time_slice[i]
            while True:
                tmp = ts.end_time - end_time
                if tmp < datetime.timedelta(0, 0, 0):
                    tmp = tmp + BaseBarHelper.one_day

                if (tmp + diff_time_span) < interval_time_span:
                    break

                end_time = end_time + (interval_time_span - diff_time_span)
                if end_time >= BaseBarHelper.one_day:
                    end_time = end_time - BaseBarHelper.one_day

                slice = TimeSlice()
                slice.begin_time = begin_time
                slice.end_time = end_time
                bars.append(slice)

                begin_time = end_time
                diff_time_span = datetime.timedelta(0, 0, 0)

            if i < (len(lst_time_slice) - 1):
                diff_time_span = diff_time_span + ts.end_time - end_time
                if diff_time_span < datetime.timedelta(0, 0, 0):
                    diff_time_span = diff_time_span + BaseBarHelper.one_day

                end_time = lst_time_slice[i + 1].begin_time

                if diff_time_span == datetime.timedelta(0, 0, 0):
                    begin_time = end_time
            else:
                if begin_time < ts.end_time:
                    slice = TimeSlice()
                    slice.begin_time = begin_time
                    slice.end_time = ts.end_time
                    bars.append(slice)

        return bars

    @staticmethod
    def create_date_time_slice(trading_day, lst_time_slices):
        """
        根据时间片段生成交日期时间片段
        :param trading_day:交易日
        :param lst_time_slices:时间片段
        :return:
        """
        pre_trading_day1 = TradingDayHelper.get_pre_trading_day(trading_day)
        pre_trading_day2 = pre_trading_day1 + datetime.timedelta(days=1)

        lst_date_time_slice = []
        for timeSlice in lst_time_slices:
            date_time_slice = DateTimeSlice()
            date_time_slice.begin_time = YfTimeHelper.join_date_time(trading_day, pre_trading_day1, pre_trading_day2, timeSlice.begin_time)
            date_time_slice.end_time = YfTimeHelper.join_date_time(trading_day, pre_trading_day1, pre_trading_day2, timeSlice.end_time)
            lst_date_time_slice.append(date_time_slice)
        return lst_date_time_slice

    @staticmethod
    def create_out_day_date_time_slice_by_date(begin_date, end_date, interval, bar_type):
        """
        非日内时间片段：据日期区间创建日内的日期时间片段
        :param begin_date:开始时间
        :param end_date:结束时间
        :param interval:间隔
        :param bar_type:K线类型
        :return:
        """
        lst_date_time_slice = []
        begin_trading_date = TradingDayHelper.get_first_trading_day(begin_date)
        end_trading_day = TradingDayHelper.get_last_trading_day(end_date)
        lst_trading_days = TradingDayHelper.get_trading_days(begin_trading_date, end_trading_day)
        end_index = len(lst_trading_days) - 1
        start_index = end_index - interval + 1
        while (end_index >= 0) & (start_index >= 0):
            begin_date = lst_trading_days[start_index]
            end_date = lst_trading_days[end_index]

            date_time_slice = DateTimeSlice()
            date_time_slice.begin_time = begin_date
            date_time_slice.end_time = end_date
            lst_date_time_slice.insert(0, date_time_slice)

            end_index = end_index - interval
            start_index = end_index - interval + 1

        if end_index >= 0:
            begin_date = lst_trading_days[0]
            end_date = lst_trading_days[end_index]

            date_time_slice = DateTimeSlice()
            date_time_slice.begin_time = begin_date
            date_time_slice.end_time = end_date
            lst_date_time_slice.insert(0, date_time_slice)

        return lst_date_time_slice

    @staticmethod
    def create_out_day_date_time_slice_by_date_time(begin_time, end_time, interval, bar_type):
        """
        非日内时间片段：据时间区间创建日内的日期时间片段
        :param begin_time:开始时间
        :param end_time:结束时间
        :param interval:间隔
        :param bar_type:K线类型
        :return:
        """
        begin_trading_date = YfTimeHelper.get_trading_day(begin_time)
        end_trading_day = YfTimeHelper.get_trading_day(end_time)
        return BaseBarHelper.create_out_day_date_time_slice_by_date(begin_trading_date, end_trading_day, interval, bar_type)

    @staticmethod
    def create_in_day_date_time_slice_by_date_time(instrument_manager, instrument_id, begin_time, end_time, interval, bar_type, *instrument_ids):
        """
        日内时间片段：根据时间区间创建日内的日期时间片段
        片段开始结束精确到时分秒
        :param instrument_manager:合约管理器
        :param instrument_id:合约编号
        :param begin_time:开始时间
        :param end_time:截止时间
        :param interval:间隔
        :param bar_type:K线类型
        :param instrument_ids:参与交易时间交集的交易品种列表
        :return:
        """
        lst_all_date_time_slices = []
        begin_trading_date = YfTimeHelper.get_trading_day(begin_time)
        end_trading_day = YfTimeHelper.get_trading_day(end_time)
        lst_trading_days = TradingDayHelper.get_trading_days(begin_trading_date, end_trading_day)
        for trading_day in lst_trading_days:
            lst_trading_time_slices = instrument_manager.get_trading_time(trading_day, instrument_id, *instrument_ids)
            lst_time_slices = BaseBarHelper.create_time_slice(lst_trading_time_slices, 0, interval, bar_type)
            lst_date_time_slices = BaseBarHelper.create_date_time_slice(trading_day, lst_time_slices)
            for dateTimeSlice in lst_date_time_slices:
                if (dateTimeSlice.end_time >= begin_time) & (dateTimeSlice.end_time <= end_time):
                    lst_all_date_time_slices.append(dateTimeSlice)

        return lst_all_date_time_slices

    @staticmethod
    def create_in_day_date_time_slice_by_date(instrument_manager, instrument_id, begin_date, end_date, interval, bar_type, *instrument_ids):
        """
        根据日期区间创建日内的日期时间片段
        :param instrument_manager:合约管理器
        :param instrument_id:合约编号
        :param begin_date:开始时间
        :param end_date:截止时间
        :param interval:间隔
        :param bar_type:K线类型
        :param instrument_ids:参与交易时间交集的交易品种列表
        :return:
        """
        lst_all_date_time_slices = []
        begin_trading_date = TradingDayHelper.get_first_trading_day(begin_date)
        end_trading_day = TradingDayHelper.get_last_trading_day(end_date)
        lst_trading_days = TradingDayHelper.get_trading_days(begin_trading_date, end_trading_day)
        for trading_day in lst_trading_days:
            lst_trading_time_slices = instrument_manager.get_trading_time(trading_day, instrument_id, *instrument_ids)
            lst_time_slices = BaseBarHelper.create_time_slice(lst_trading_time_slices, 0, interval, bar_type)
            lst_date_time_slices = BaseBarHelper.create_date_time_slice(trading_day, lst_time_slices)
            for dateTimeSlice in lst_date_time_slices:
                lst_all_date_time_slices.append(dateTimeSlice)

        return lst_all_date_time_slices

    @staticmethod
    def create_in_day_date_time_slice_by_date_consider_holiday(instrument_manager, instrument_id, begin_date, end_date, interval, bar_type, *instrument_ids):
        """
        根据日期区间创建日内的日期时间片段
        :param instrument_manager:合约管理器
        :param instrument_id:合约编号
        :param begin_date:开始时间
        :param end_date:截止时间
        :param interval:间隔
        :param bar_type:K线类型
        :param instrument_ids:参与交易时间交集的交易品种列表
        :return:
        """
        lst_all_date_time_slices = []
        begin_trading_date = TradingDayHelper.get_first_trading_day(begin_date)
        end_trading_day = TradingDayHelper.get_last_trading_day(end_date)
        lst_trading_days = TradingDayHelper.get_trading_days(begin_trading_date, end_trading_day)
        listing_date = instrument_manager.get_listing_date(EnumMarket.期货, instrument_id)
        for trading_day in lst_trading_days:
            lst_trading_time_slices = instrument_manager.get_trading_time(trading_day, instrument_id, *instrument_ids)
            lst_time_slices = BaseBarHelper.create_time_slice(lst_trading_time_slices, 0, interval, bar_type)
            lst_date_time_slices = BaseBarHelper.create_date_time_slice(trading_day, lst_time_slices)
            for dateTimeSlice in lst_date_time_slices:
                if trading_day == listing_date or HolidayHelper.is_post_holiday_trading_date(trading_day):
                    if dateTimeSlice.begin_time.hour > 6 and (dateTimeSlice.begin_time.hour < 18):  # 只有日盘
                        lst_all_date_time_slices.append(dateTimeSlice)
                else:
                    lst_all_date_time_slices.append(dateTimeSlice)

        return lst_all_date_time_slices

    @staticmethod
    def create_data_frame_in_day_date_time_slice_by_date_consider_holiday(instrument_manager, instrument_id, begin_date, end_date, interval, bar_type, *instrument_ids):
        """
        根据日期区间创建日内的日期时间片段的DataFrame
        :param instrument_manager:合约管理器
        :param instrument_id:合约编号
        :param begin_date:开始时间
        :param end_date:截止时间
        :param interval:间隔
        :param bar_type:K线类型
        :param instrument_ids:参与交易时间交集的交易品种列表
        :return:
        """
        lst_all_date_time_slices = []
        begin_trading_date = TradingDayHelper.get_first_trading_day(begin_date)
        end_trading_day = TradingDayHelper.get_last_trading_day(end_date)
        lst_trading_days = TradingDayHelper.get_trading_days(begin_trading_date, end_trading_day)
        listing_date = instrument_manager.get_listing_date(EnumMarket.期货, instrument_id)
        for trading_day in lst_trading_days:
            lst_trading_time_slices = instrument_manager.get_trading_time(trading_day, instrument_id, *instrument_ids)
            lst_time_slices = BaseBarHelper.create_time_slice(lst_trading_time_slices, 0, interval, bar_type)
            lst_date_time_slices = BaseBarHelper.create_date_time_slice(trading_day, lst_time_slices)
            for dateTimeSlice in lst_date_time_slices:
                if trading_day == listing_date or HolidayHelper.is_post_holiday_trading_date(trading_day):
                    if dateTimeSlice.begin_time.hour > 6 and (dateTimeSlice.begin_time.hour < 18):  # 只有日盘
                        lst_all_date_time_slices.append([dateTimeSlice.begin_time, dateTimeSlice.end_time])
                else:
                    lst_all_date_time_slices.append([dateTimeSlice.begin_time, dateTimeSlice.end_time])

        df_all_date_time_slices = pd.DataFrame(lst_all_date_time_slices, columns=['begin_time', 'end_time'])
        return df_all_date_time_slices

    @staticmethod
    def create_night_am_pm_date_time_slice_by_date(instrument_manager, instrument_id, begin_date, end_date, *instrument_ids):
        """
        根据日期区间创建日内的日期时间片段
        :param instrument_manager:合约管理器
        :param instrument_id:合约编号
        :param begin_date:开始时间
        :param end_date:截止时间
        :param instrument_ids:参与交易时间交集的交易品种列表
        :return:
        """
        lst_all_date_time_slices = []
        begin_trading_date = TradingDayHelper.get_first_trading_day(begin_date)
        end_trading_day = TradingDayHelper.get_last_trading_day(end_date)
        lst_trading_days = TradingDayHelper.get_trading_days(begin_trading_date, end_trading_day)
        for trading_day in lst_trading_days:
            lst_trading_time_slices = instrument_manager.get_trading_time(trading_day, instrument_id, *instrument_ids)
            lst_time_slices = []
            for i in range(len(lst_trading_time_slices)):
                if 10 <= (lst_trading_time_slices[i].begin_time.seconds / (60 * 60)) <= 11:
                    lst_time_slices[i - 1].end_time = lst_trading_time_slices[i].end_time
                else:
                    lst_time_slices.append(lst_trading_time_slices[i])
            lst_date_time_slices = BaseBarHelper.create_date_time_slice(trading_day, lst_time_slices)
            for dateTimeSlice in lst_date_time_slices:
                lst_all_date_time_slices.append(dateTimeSlice)

        return lst_all_date_time_slices

    @staticmethod
    def create_one_day_date_time_slice(instrument_manager, instrument_id, trading_day, interval, bar_type, *instrument_ids):
        """
        创建一天的日期时间片段
        :param instrument_manager:合约管理器
        :param instrument_id:合约编号
        :param trading_day:交易日
        :param interval:间隔
        :param bar_type:K线类型
        :param instrument_ids:参与交易时间交集的交易品种列表
        :return:
        """
        lst_trading_time_slices = instrument_manager.get_trading_time(trading_day, instrument_id, *instrument_ids)
        lst_time_slices = BaseBarHelper.create_time_slice(lst_trading_time_slices, 0, interval, bar_type)
        lst_date_time_slices = BaseBarHelper.create_date_time_slice(trading_day, lst_time_slices)
        return lst_date_time_slices



class PositionKey:
    """
    持仓唯一索引
    """
    def __init__(self, product_code, strategy_id, instrument_id, direction):
        self.product_code = product_code
        self.strategy_id = strategy_id
        self.instrument_id = instrument_id
        self.direction = direction

    def equals(self, obj):
        """

        :param obj:
        :return:
        """
        if not isinstance(obj, PositionKey):
            return False

        return (self.product_code == obj.product_code) & (self.strategy_id == obj.strategy_id) & (self.instrument_id == obj.instrument_id) & (
                self.direction == obj.direction)

    def get_hash_code(self):
        """

        :return:
        """
        return '{0}_{1}_{2}_{3}'.format(self.product_code, self.strategy_id, self.instrument_id, self.direction)

    def to_string(self):
        """

        :return:
        """
        return '{0}_{1}_{2}_{3}'.format(self.product_code, self.strategy_id, self.instrument_id, self.direction)


from SapphireQuant.Core.Entity import Entity
from SapphireQuant.Core.Enum import EnumPositionType, EnumPositionDirection, EnumHedgeFlag


class FuturePositionDetail(Entity):
    """
    期货持仓
    """

    def __init__(self):
        super().__init__()
        self.instrument_id = ''
        self.broker_id = ''
        self.investor_id = ''
        self.instrument_name = ''
        self.direction = EnumPositionDirection.净持仓
        self.hedge_flag = EnumHedgeFlag.套保
        self.exchange_id = ''
        self.open_time = None
        self.trade_sys_id = ''
        self.open_price = 0
        self.volume = 0
        self.float_profit = 0
        self.position_profit = 0
        self.market_value = 0
        self.position_type = EnumPositionType.今仓

    def clone(self):
        position = FuturePositionDetail()
        position.product_code = self.product_code
        position.strategy_id = self.strategy_id
        position.instrument_id = self.instrument_id
        position.broker_id = self.broker_id
        position.investor_id = self.investor_id
        position.instrument_name = self.instrument_name
        position.direction = self.direction
        position.hedge_flag = self.hedge_flag
        position.exchange_id = self.exchange_id
        position.open_time = self.open_time
        position.open_price = self.open_price
        position.volume = self.volume
        position.float_profit = self.float_profit
        position.position_profit = self.position_profit
        position.market_value = self.market_value

        return position

    def to_string(self):
        """

        :return:
        """
        return '{0}:{1},成交编号:{2},手数:{3},开仓价:{4},持仓类型:{5},开仓时间:{6},持仓盈亏:{7},浮动盈亏:{8},持仓市值:{9}'.format(
            self.instrument_id, self.direction.chinese_name(), self.trade_sys_id, self.volume, self.open_price, self.position_type.chinese_name(),
            self.open_time.strftime('%Y%m%d'), self.position_profit, self.float_profit, self.market_value)



from .FutureCache import FutureCache
from .FutureData import FutureData
from .OldFutureCache import OldFutureCache
from .PositionKey import PositionKey


__all__ = [
    'FutureCache',
    'FutureData',
    'PositionKey',
    'OldFutureCache'
]


from SapphireQuant.Config.TradingFrame.TradingFrameHelper import TradingFrameHelper


class TradingFrameNode:
    def __init__(self):
        self._begin_day_string = ""
        self._trading_time = ""

    @property
    def begin_day_string(self):
        """

        :return:
        """
        return self._begin_day_string

    @begin_day_string.setter
    def begin_day_string(self, value):
        self._begin_day_string = value

    @property
    def trading_time(self):
        """

        :return:
        """
        return self._trading_time

    @trading_time.setter
    def trading_time(self, value):
        self._trading_time = value

    @property
    def begin_day(self):
        """

        :return:
        """
        return TradingFrameHelper.parse_date(self._begin_day_string)

    @property
    def trading_time_slices(self):
        """

        :return:
        """
        return TradingFrameHelper.parse_time_slice(self._trading_time)
